# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Fake LDAP server for test harness.

This class does very little error checking, and knows nothing about ldap
class definitions.  It implements the minimum emulation of the python ldap
library to work with keystone.

"""

import random
import re
import shelve

import ldap
from oslo_log import log

import keystone.conf
from keystone import exception
from keystone.identity.backends.ldap import common

SCOPE_NAMES = {
    ldap.SCOPE_BASE: 'SCOPE_BASE',
    ldap.SCOPE_ONELEVEL: 'SCOPE_ONELEVEL',
    ldap.SCOPE_SUBTREE: 'SCOPE_SUBTREE',
}

LOG = log.getLogger(__name__)
CONF = keystone.conf.CONF


def _internal_attr(attr_name, value_or_values):
    def normalize_value(value):
        return common.utf8_decode(value)

    def normalize_dn(dn):
        # Capitalize the attribute names as an LDAP server might.

        # NOTE(blk-u): Special case for this tested value, used with
        # test_user_id_comma. The call to str2dn here isn't always correct
        # here, because `dn` is escaped for an LDAP filter. str2dn() normally
        # works only because there's no special characters in `dn`.
        if dn == 'cn=Doe\\5c, John,ou=Users,cn=example,cn=com':
            return 'CN=Doe\\, John,OU=Users,CN=example,CN=com'

        # NOTE(blk-u): Another special case for this tested value. When a
        # roleOccupant has an escaped comma, it gets converted to \2C.
        if dn == 'cn=Doe\\, John,ou=Users,cn=example,cn=com':
            return 'CN=Doe\\2C John,OU=Users,CN=example,CN=com'

        try:
            dn = ldap.dn.str2dn(dn)
        except ldap.DECODING_ERROR:
            # NOTE(amakarov): In case of IDs instead of DNs in group members
            # they must be handled as regular values.
            return normalize_value(dn)

        norm = []
        for part in dn:
            name, val, i = part[0]
            name = name.upper()
            norm.append([(name, val, i)])
        return ldap.dn.dn2str(norm)

    if attr_name in ('member', 'roleOccupant'):
        attr_fn = normalize_dn
    else:
        attr_fn = normalize_value

    if isinstance(value_or_values, list):
        return [attr_fn(x) for x in value_or_values]
    return [attr_fn(value_or_values)]


def _match_query(query, attrs, attrs_checked):
    """Match an ldap query to an attribute dictionary.

    The characters &, |, and ! are supported in the query. No syntax checking
    is performed, so malformed queries will not work correctly.
    """
    # cut off the parentheses
    inner = query[1:-1]
    if inner.startswith(('&', '|')):
        if inner[0] == '&':
            matchfn = all
        else:
            matchfn = any
        # cut off the & or |
        groups = _paren_groups(inner[1:])
        return matchfn(
            _match_query(group, attrs, attrs_checked) for group in groups
        )
    if inner.startswith('!'):
        # cut off the ! and the nested parentheses
        return not _match_query(query[2:-1], attrs, attrs_checked)

    (k, _sep, v) = inner.partition('=')
    attrs_checked.add(k.lower())
    return _match(k, v, attrs)


def _paren_groups(source):
    """Split a string into parenthesized groups."""
    count = 0
    start = 0
    result = []
    for pos in range(len(source)):
        if source[pos] == '(':
            if count == 0:
                start = pos
            count += 1
        if source[pos] == ')':
            count -= 1
            if count == 0:
                result.append(source[start : pos + 1])
    return result


def _match(key, value, attrs):
    """Match a given key and value against an attribute list."""

    def match_with_wildcards(norm_val, val_list):
        # Case insensitive checking with wildcards
        if norm_val.startswith('*'):
            if norm_val.endswith('*'):
                # Is the string anywhere in the target?
                for x in val_list:
                    if norm_val[1:-1] in x:
                        return True
            else:
                # Is the string at the end of the target?
                for x in val_list:
                    if norm_val[1:] == x[len(x) - len(norm_val) + 1 :]:
                        return True
        elif norm_val.endswith('*'):
            # Is the string at the start of the target?
            for x in val_list:
                if norm_val[:-1] == x[: len(norm_val) - 1]:
                    return True
        else:
            # Is the string an exact match?
            for x in val_list:
                if check_value == x:
                    return True
        return False

    if key not in attrs:
        return False
    # This is a pure wild card search, so the answer must be yes!
    if value == '*':
        return True
    if key == 'serviceId':
        # For serviceId, the backend is returning a list of numbers.
        # Make sure we convert them to strings first before comparing
        # them.
        str_sids = [str(x) for x in attrs[key]]
        return str(value) in str_sids
    if key != 'objectclass':
        check_value = _internal_attr(key, value)[0].lower()
        norm_values = [_internal_attr(key, x)[0].lower() for x in attrs[key]]
        return match_with_wildcards(check_value, norm_values)
    # It is an objectclass check, so check subclasses
    values = _subs(value)
    for v in values:
        if v in attrs[key]:
            return True
    return False


def _subs(value):
    """Return a list of subclass strings.

    The strings represent the ldap objectclass plus any subclasses that
    inherit from it. Fakeldap doesn't know about the ldap object structure,
    so subclasses need to be defined manually in the dictionary below.

    """
    subs = {
        'groupOfNames': [
            'keystoneProject',
            'keystoneRole',
            'keystoneProjectRole',
        ]
    }
    if value in subs:
        return [value] + subs[value]
    return [value]


server_fail = False


class FakeShelve(dict):
    def sync(self):
        pass


FakeShelves = {}
PendingRequests = {}


class FakeLdap(common.LDAPHandler):
    """Emulate the python-ldap API.

    The python-ldap API requires all strings to be UTF-8 encoded with the
    exception of [1]. This is assured by the caller of this interface
    (i.e. KeystoneLDAPHandler).

    However, internally this emulation MUST process and store strings
    in a canonical form which permits operations on
    characters. Encoded strings do not provide the ability to operate
    on characters. Therefore this emulation accepts UTF-8 encoded
    strings, decodes them to unicode for operations internal to this
    emulation, and encodes them back to UTF-8 when returning values
    from the emulation.

    [1] Some fields (DNs, RDNs, attribute names, queries) are represented
    as text in python-ldap for Python 3, and for Python 2 when
    bytes_mode=False. For more details see:
    http://www.python-ldap.org/en/latest/bytes_mode.html#bytes-mode

    """

    __prefix = 'ldap:'

    def __init__(self, conn=None):
        super().__init__(conn=conn)
        self._ldap_options = {ldap.OPT_DEREF: ldap.DEREF_NEVER}

    def connect(
        self,
        url,
        page_size=0,
        alias_dereferencing=None,
        use_tls=False,
        tls_cacertfile=None,
        tls_cacertdir=None,
        tls_req_cert='demand',
        chase_referrals=None,
        debug_level=None,
        use_pool=None,
        pool_size=None,
        pool_retry_max=None,
        pool_retry_delay=None,
        pool_conn_timeout=None,
        pool_conn_lifetime=None,
        conn_timeout=None,
    ):
        if url.startswith('fake://memory'):
            if url not in FakeShelves:
                FakeShelves[url] = FakeShelve()
            self.db = FakeShelves[url]
        else:
            self.db = shelve.open(url[7:])

        using_ldaps = url.lower().startswith("ldaps")

        if use_tls and using_ldaps:
            raise AssertionError('Invalid TLS / LDAPS combination')

        if use_tls:
            if tls_cacertfile:
                ldap.set_option(ldap.OPT_X_TLS_CACERTFILE, tls_cacertfile)
            elif tls_cacertdir:
                ldap.set_option(ldap.OPT_X_TLS_CACERTDIR, tls_cacertdir)
            if tls_req_cert in list(common.LDAP_TLS_CERTS.values()):
                ldap.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, tls_req_cert)
            else:
                raise ValueError(
                    "invalid TLS_REQUIRE_CERT tls_req_cert=%s", tls_req_cert
                )

        if alias_dereferencing is not None:
            self.set_option(ldap.OPT_DEREF, alias_dereferencing)
        self.page_size = page_size

        self.use_pool = use_pool
        self.pool_size = pool_size
        self.pool_retry_max = pool_retry_max
        self.pool_retry_delay = pool_retry_delay
        self.pool_conn_timeout = pool_conn_timeout
        self.pool_conn_lifetime = pool_conn_lifetime
        self.conn_timeout = conn_timeout

    def _dn_to_id_attr(self, dn):
        return ldap.dn.str2dn(dn)[0][0][0]

    def _dn_to_id_value(self, dn):
        return ldap.dn.str2dn(dn)[0][0][1]

    def key(self, dn):
        return f'{self.__prefix}{dn}'

    def simple_bind_s(
        self, who='', cred='', serverctrls=None, clientctrls=None
    ):
        """Provide for compatibility but this method is ignored."""
        if server_fail:
            raise ldap.SERVER_DOWN
        whos = ['cn=Admin', CONF.ldap.user]
        if who in whos and cred in ['password', CONF.ldap.password]:
            self.connected = True
            self.who = who
            self.cred = cred
            return

        attrs = self.db.get(self.key(who))
        if not attrs:
            LOG.debug('who=%s not found, binding anonymously', who)

        db_password = ''
        if attrs:
            try:
                db_password = attrs['userPassword'][0]
            except (KeyError, IndexError):
                LOG.debug('bind fail: password for who=%s not found', who)
                raise ldap.INAPPROPRIATE_AUTH

        if cred != db_password:
            LOG.debug('bind fail: password for who=%s does not match', who)
            raise ldap.INVALID_CREDENTIALS

    def unbind_s(self):
        """Provide for compatibility but this method is ignored."""
        self.connected = False
        self.who = None
        self.cred = None
        if server_fail:
            raise ldap.SERVER_DOWN

    def add_s(self, dn, modlist):
        """Add an object with the specified attributes at dn."""
        if server_fail:
            raise ldap.SERVER_DOWN

        id_attr_in_modlist = False
        id_attr = self._dn_to_id_attr(dn)
        id_value = self._dn_to_id_value(dn)

        # The LDAP API raises a TypeError if attr name is None.
        for k, dummy_v in modlist:
            if k is None:
                raise TypeError(f'must be string, not None. modlist={modlist}')

            if k == id_attr:
                for val in dummy_v:
                    if common.utf8_decode(val) == id_value:
                        id_attr_in_modlist = True

        if not id_attr_in_modlist:
            LOG.debug(
                'id_attribute=%(attr)s missing, attributes=%(attrs)s',
                {'attr': id_attr, 'attrs': modlist},
            )
            raise ldap.NAMING_VIOLATION
        key = self.key(dn)
        LOG.debug(
            'add item: dn=%(dn)s, attrs=%(attrs)s',
            {'dn': dn, 'attrs': modlist},
        )
        if key in self.db:
            LOG.debug('add item failed: dn=%s is already in store.', dn)
            raise ldap.ALREADY_EXISTS(dn)

        self.db[key] = {k: _internal_attr(k, v) for k, v in modlist}
        self.db.sync()

    def delete_s(self, dn):
        """Remove the ldap object at specified dn."""
        return self.delete_ext_s(dn, serverctrls=[])

    def _getChildren(self, dn):
        return [
            k
            for k, v in self.db.items()
            if re.match(f'{re.escape(self.__prefix)}.*,{re.escape(dn)}', k)
        ]

    def delete_ext_s(self, dn, serverctrls, clientctrls=None):
        """Remove the ldap object at specified dn."""
        if server_fail:
            raise ldap.SERVER_DOWN

        try:
            key = self.key(dn)
            LOG.debug('FakeLdap delete item: dn=%s', dn)
            del self.db[key]
        except KeyError:
            LOG.debug('delete item failed: dn=%s not found.', dn)
            raise ldap.NO_SUCH_OBJECT
        self.db.sync()

    def modify_s(self, dn, modlist):
        """Modify the object at dn using the attribute list.

        :param dn: an LDAP DN
        :param modlist: a list of tuples in the following form:
                      ([MOD_ADD | MOD_DELETE | MOD_REPACE], attribute, value)
        """
        if server_fail:
            raise ldap.SERVER_DOWN

        key = self.key(dn)
        LOG.debug(
            'modify item: dn=%(dn)s attrs=%(attrs)s',
            {'dn': dn, 'attrs': modlist},
        )
        try:
            entry = self.db[key]
        except KeyError:
            LOG.debug('modify item failed: dn=%s not found.', dn)
            raise ldap.NO_SUCH_OBJECT

        for cmd, k, v in modlist:
            values = entry.setdefault(k, [])
            if cmd == ldap.MOD_ADD:
                v = _internal_attr(k, v)
                for x in v:
                    if x in values:
                        raise ldap.TYPE_OR_VALUE_EXISTS
                values += v
            elif cmd == ldap.MOD_REPLACE:
                values[:] = _internal_attr(k, v)
            elif cmd == ldap.MOD_DELETE:
                if v is None:
                    if not values:
                        LOG.debug(
                            'modify item failed: '
                            'item has no attribute "%s" to delete',
                            k,
                        )
                        raise ldap.NO_SUCH_ATTRIBUTE
                    values[:] = []
                else:
                    for val in _internal_attr(k, v):
                        try:
                            values.remove(val)
                        except ValueError:
                            LOG.debug(
                                'modify item failed: '
                                'item has no attribute "%(k)s" with '
                                'value "%(v)s" to delete',
                                {'k': k, 'v': val},
                            )
                            raise ldap.NO_SUCH_ATTRIBUTE
            else:
                LOG.debug('modify item failed: unknown command %s', cmd)
                raise NotImplementedError(
                    f'modify_s action {cmd} not implemented'
                )
        self.db[key] = entry
        self.db.sync()

    def search_s(
        self,
        base,
        scope,
        filterstr='(objectClass=*)',
        attrlist=None,
        attrsonly=0,
    ):
        """Search for all matching objects under base using the query.

        Args:
        base -- dn to search under
        scope -- search scope (base, subtree, onelevel)
        filterstr -- filter objects by
        attrlist -- attrs to return. Returns all attrs if not specified

        """
        if server_fail:
            raise ldap.SERVER_DOWN

        if (not filterstr) and (scope != ldap.SCOPE_BASE):
            raise AssertionError(
                'Search without filter on onelevel or subtree scope'
            )

        if scope == ldap.SCOPE_BASE:
            try:
                item_dict = self.db[self.key(base)]
            except KeyError:
                LOG.debug('search fail: dn not found for SCOPE_BASE')
                raise ldap.NO_SUCH_OBJECT
            results = [(base, item_dict)]
        elif scope == ldap.SCOPE_SUBTREE:
            # FIXME - LDAP search with SUBTREE scope must return the base
            # entry, but the code below does _not_. Unfortunately, there are
            # several tests that depend on this broken behavior, and fail
            # when the base entry is returned in the search results. The
            # fix is easy here, just initialize results as above for
            # the SCOPE_BASE case.
            # https://bugs.launchpad.net/keystone/+bug/1368772
            try:
                item_dict = self.db[self.key(base)]
            except KeyError:
                LOG.debug('search fail: dn not found for SCOPE_SUBTREE')
                raise ldap.NO_SUCH_OBJECT
            results = [(base, item_dict)]
            extraresults = [
                (k[len(self.__prefix) :], v)
                for k, v in self.db.items()
                if re.match(
                    f'{re.escape(self.__prefix)}.*,{re.escape(base)}', k
                )
            ]
            results.extend(extraresults)
        elif scope == ldap.SCOPE_ONELEVEL:

            def get_entries():
                base_dn = ldap.dn.str2dn(base)
                base_len = len(base_dn)

                for k, v in self.db.items():
                    if not k.startswith(self.__prefix):
                        continue
                    k_dn_str = k[len(self.__prefix) :]
                    k_dn = ldap.dn.str2dn(k_dn_str)
                    if len(k_dn) != base_len + 1:
                        continue
                    if k_dn[-base_len:] != base_dn:
                        continue
                    yield (k_dn_str, v)

            results = list(get_entries())

        else:
            # openldap client/server raises PROTOCOL_ERROR for unexpected scope
            raise ldap.PROTOCOL_ERROR

        objects = []
        for dn, attrs in results:
            # filter the objects by filterstr
            id_attr, id_val, _ = ldap.dn.str2dn(dn)[0][0]
            match_attrs = attrs.copy()
            match_attrs[id_attr] = [id_val]
            attrs_checked = set()
            if not filterstr or _match_query(
                filterstr, match_attrs, attrs_checked
            ):
                if (
                    filterstr
                    and (scope != ldap.SCOPE_BASE)
                    and ('objectclass' not in attrs_checked)
                ):
                    raise AssertionError('No objectClass in search filter')
                # filter the attributes by attrlist
                attrs = {
                    k: v
                    for k, v in attrs.items()
                    if not attrlist or k in attrlist
                }
                objects.append((dn, attrs))

        return objects

    def set_option(self, option, invalue):
        self._ldap_options[option] = invalue

    def get_option(self, option):
        value = self._ldap_options.get(option)
        return value

    def search_ext(
        self,
        base,
        scope,
        filterstr='(objectClass=*)',
        attrlist=None,
        attrsonly=0,
        serverctrls=None,
        clientctrls=None,
        timeout=-1,
        sizelimit=0,
    ):
        if clientctrls is not None or timeout != -1 or sizelimit != 0:
            raise exception.NotImplemented()

        # only passing a single server control is supported by this fake ldap
        if serverctrls and len(serverctrls) > 1:
            raise exception.NotImplemented()

        # search_ext is async and returns an identifier used for
        # retrieving the results via result3(). This will be emulated by
        # storing the request in a variable with random integer key and
        # performing the real lookup in result3()
        msgid = random.randint(0, 1000)
        PendingRequests[msgid] = (
            base,
            scope,
            filterstr,
            attrlist,
            attrsonly,
            serverctrls,
        )
        return msgid

    def result3(
        self, msgid=ldap.RES_ANY, all=1, timeout=None, resp_ctrl_classes=None
    ):
        """Execute async request.

        Only msgid param is supported. Request info is fetched from global
        variable `PendingRequests` by msgid, executed using search_s and
        limited if requested.
        """
        if all != 1 or timeout is not None or resp_ctrl_classes is not None:
            raise exception.NotImplemented()

        params = PendingRequests[msgid]
        # search_s accepts a subset of parameters of search_ext,
        # that's why we use only the first 5.
        results = self.search_s(*params[:5])

        # extract limit from serverctrl
        serverctrls = params[5]
        ctrl = serverctrls[0]

        if ctrl.size:
            rdata = results[: ctrl.size]
        else:
            rdata = results

        # real result3 returns various service info -- rtype, rmsgid,
        # serverctrls. Now this info is not used, so all this info is None
        rtype = None
        rmsgid = None
        serverctrls = None
        return (rtype, rdata, rmsgid, serverctrls)


class FakeLdapPool(FakeLdap):
    """Emulate the python-ldap API with pooled connections.

    This class is used as connector class in PooledLDAPHandler.

    """

    def __init__(self, uri, retry_max=None, retry_delay=None, conn=None):
        super().__init__(conn=conn)
        self.url = uri
        self._uri = uri
        self.connected = None
        self.conn = self
        self._connection_time = 5  # any number greater than 0

    def get_lifetime(self):
        return self._connection_time

    def simple_bind_s(
        self, who=None, cred=None, serverctrls=None, clientctrls=None
    ):
        if self.url.startswith('fakepool://memory'):
            if self.url not in FakeShelves:
                FakeShelves[self.url] = FakeShelve()
            self.db = FakeShelves[self.url]
        else:
            self.db = shelve.open(self.url[11:])

        if not who:
            who = 'cn=Admin'
        if not cred:
            cred = 'password'

        super().simple_bind_s(
            who=who,
            cred=cred,
            serverctrls=serverctrls,
            clientctrls=clientctrls,
        )

    def unbind_ext_s(self):
        """Added to extend FakeLdap as connector class."""
        pass


class FakeLdapNoSubtreeDelete(FakeLdap):
    """FakeLdap subclass that does not support subtree delete.

    Same as FakeLdap except delete will throw the LDAP error
    ldap.NOT_ALLOWED_ON_NONLEAF if there is an attempt to delete
    an entry that has children.
    """

    def delete_ext_s(self, dn, serverctrls, clientctrls=None):
        """Remove the ldap object at specified dn."""
        if server_fail:
            raise ldap.SERVER_DOWN

        try:
            children = self._getChildren(dn)
            if children:
                raise ldap.NOT_ALLOWED_ON_NONLEAF

        except KeyError:
            LOG.debug('delete item failed: dn=%s not found.', dn)
            raise ldap.NO_SUCH_OBJECT
        super().delete_ext_s(dn, serverctrls, clientctrls)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import copy
import uuid

from keystone.exception import ValidationError
from keystone.federation import utils
from keystone.tests import unit


class TestFederationUtils(unit.TestCase):
    def setUp(self):
        super().setUp()
        self.mapping_id_mock = uuid.uuid4().hex
        self.domain_id_mock = uuid.uuid4().hex
        self.domain_mock = {'id': self.domain_id_mock}
        self.attribute_mapping_schema_1_0 = {
            "id": self.mapping_id_mock,
            "schema_version": '1.0',
            "rules": [
                {
                    "remote": [
                        {"type": "OIDC-preferred_username"},
                        {"type": "OIDC-email"},
                        {"type": "OIDC-openstack-user-domain"},
                        {"type": "OIDC-openstack-default-project"},
                        {
                            "type": "OIDC-openstack-user-status",
                            "any_one_of": ["local"],
                        },
                    ],
                    "local": [
                        {
                            "domain": {"name": "{2}"},
                            "user": {
                                "domain": {"name": "{2}"},
                                "type": "local",
                                "name": "{0}",
                                "email": "{1}",
                            },
                            "projects": [
                                {"name": "{3}", "roles": [{"name": "member"}]}
                            ],
                        }
                    ],
                }
            ],
        }
        self.attribute_mapping_schema_2_0 = copy.deepcopy(
            self.attribute_mapping_schema_1_0
        )
        self.attribute_mapping_schema_2_0['schema_version'] = '2.0'
        self.attribute_mapping_schema_2_0['rules'][0]['local'][0]["projects"][
            0
        ]['domain'] = {"name": "{some_place_holder}"}
        self.rule_processor = utils.RuleProcessor(
            self.mapping_id_mock, self.attribute_mapping_schema_1_0
        )
        self.rule_processor_schema_2_0 = (
            utils.RuleProcessorToHonorDomainOption(
                self.mapping_id_mock, self.attribute_mapping_schema_2_0
            )
        )

    def test_validate_mapping_structure_schema1_0(self):
        utils.validate_mapping_structure(self.attribute_mapping_schema_1_0)

    def test_validate_mapping_structure_schema2_0(self):
        utils.validate_mapping_structure(self.attribute_mapping_schema_2_0)

    def test_normalize_user_no_type_set(self):
        user = {}
        self.rule_processor.normalize_user(user, self.domain_mock)
        self.assertEqual(utils.UserType.EPHEMERAL, user['type'])

    def test_normalize_user_unexpected_type(self):
        user = {'type': "weird-type"}
        self.assertRaises(
            ValidationError,
            self.rule_processor.normalize_user,
            user,
            self.domain_mock,
        )

    def test_normalize_user_type_local(self):
        user = {'type': utils.UserType.LOCAL}
        self.rule_processor.normalize_user(user, self.domain_mock)
        self.assertEqual(utils.UserType.LOCAL, user['type'])

    def test_normalize_user_type_ephemeral(self):
        user = {'type': utils.UserType.EPHEMERAL}
        self.rule_processor.normalize_user(user, self.domain_mock)
        self.assertEqual(utils.UserType.EPHEMERAL, user['type'])

    def test_extract_groups(self):
        group1 = {'name': "group1", 'domain': self.domain_id_mock}
        group_by_domain = {self.domain_id_mock: [group1]}

        result = utils.RuleProcessor(
            self.mapping_id_mock, self.attribute_mapping_schema_1_0
        ).extract_groups(group_by_domain)

        self.assertEqual([group1], list(result))

    def test_process_group_by_name_domain_with_name_only(self):
        domain = {'name': "domain1"}
        group1 = {'name': "group1", 'domain': domain}
        group_by_domain = {}
        result = self.rule_processor.process_group_by_name(
            group1, group_by_domain
        )
        self.assertEqual([group1], list(result))
        self.assertEqual([domain["name"]], list(group_by_domain.keys()))

    def test_process_group_by_name_domain_with_id_only(self):
        group1 = {'name': "group1", 'domain': self.domain_mock}
        group_by_domain = {}
        result = self.rule_processor.process_group_by_name(
            group1, group_by_domain
        )
        self.assertEqual([group1], list(result))
        self.assertEqual([self.domain_id_mock], list(group_by_domain.keys()))

    def test_process_group_by_name_domain_with_id_and_name(self):
        self.domain_mock['name'] = "domain1"
        group1 = {'name': "group1", 'domain': self.domain_mock}
        group_by_domain = {}
        result = self.rule_processor.process_group_by_name(
            group1, group_by_domain
        )
        self.assertEqual([group1], list(result))
        self.assertEqual(["domain1"], list(group_by_domain.keys()))

    def test_process_group_by_name_groups_same_domain(self):
        group1 = {'name': "group1", 'domain': self.domain_mock}
        group2 = {'name': "group2", 'domain': self.domain_mock}
        group_by_domain = {self.domain_id_mock: [group1]}
        result = self.rule_processor.process_group_by_name(
            group2, group_by_domain
        )
        self.assertEqual([group1, group2], list(result))
        self.assertEqual([self.domain_id_mock], list(group_by_domain.keys()))

    def test_process_group_by_name_groups_different_domain(self):
        domain = {'name': "domain1"}
        group1 = {'name': "group1", 'domain': domain}
        group2 = {'name': "group2", 'domain': self.domain_mock}
        group_by_domain = {"domain1": [group1]}
        result = self.rule_processor.process_group_by_name(
            group2, group_by_domain
        )
        self.assertEqual([group1, group2], list(result))
        self.assertEqual(
            ["domain1", self.domain_id_mock], list(group_by_domain.keys())
        )

    def test_rule_processor_extract_projects_schema1_0_no_projects(self):
        result = self.rule_processor.extract_projects({})
        self.assertEqual([], result)

    def test_rule_processor_extract_projects_schema1_0(self):
        projects_list = [{'name': "project1", 'domain': self.domain_mock}]
        identity_values = {'projects': projects_list}
        result = self.rule_processor.extract_projects(identity_values)
        self.assertEqual(projects_list, result)

    def test_rule_processor_extract_projects_schema2_0_no_projects(self):
        result = self.rule_processor_schema_2_0.extract_projects({})
        self.assertEqual([], result)

    def test_rule_processor_extract_projects_schema2_0_domain_in_project(self):
        projects_list = [{'name': "project1", 'domain': self.domain_mock}]
        identity_values = {'projects': projects_list}
        result = self.rule_processor_schema_2_0.extract_projects(
            identity_values
        )
        self.assertEqual(projects_list, result)

    def test_rule_processor_extract_projects_schema2_0_no_domain(self):
        projects_list = [{'name': "project1"}]
        identity_values = {'projects': projects_list}
        result = self.rule_processor_schema_2_0.extract_projects(
            identity_values
        )
        self.assertEqual(projects_list, result)

    def test_rule_processor_extract_projects_schema2_0_no_domain_project(self):
        project = {'name': "project1"}
        identity_values = {
            'projects': [project.copy()],
            'domain': self.domain_mock,
        }
        result = self.rule_processor_schema_2_0.extract_projects(
            identity_values
        )
        expected_project = project.copy()
        expected_project['domain'] = self.domain_mock
        self.assertEqual([expected_project], result)

    def test_normalize_user_no_type_set_schema_2_0(self):
        user = {}
        self.rule_processor_schema_2_0.normalize_user(user, self.domain_mock)
        self.assertEqual(utils.UserType.EPHEMERAL, user['type'])

    def test_normalize_user_unexpected_type_schema_2_0(self):
        user = {'type': "weird-type"}
        self.assertRaises(
            ValidationError,
            self.rule_processor_schema_2_0.normalize_user,
            user,
            self.domain_mock,
        )

    def test_normalize_user_type_local_schema_2_0(self):
        user = {'type': utils.UserType.LOCAL}
        self.rule_processor_schema_2_0.normalize_user(user, self.domain_mock)
        self.assertEqual(utils.UserType.LOCAL, user['type'])

    def test_normalize_user_type_ephemeral_schema_2_0(self):
        user = {'type': utils.UserType.EPHEMERAL}
        self.rule_processor_schema_2_0.normalize_user(user, self.domain_mock)
        self.assertEqual(utils.UserType.EPHEMERAL, user['type'])

    def test_normalize_user_no_domain_schema_2_0(self):
        user = {}
        self.rule_processor_schema_2_0.normalize_user(user, self.domain_mock)
        self.assertEqual(utils.UserType.EPHEMERAL, user['type'])
        self.assertEqual(self.domain_mock, user.get("domain"))

    def test_create_attribute_mapping_rules_processor_default(self):
        result = utils.create_attribute_mapping_rules_processor(
            self.attribute_mapping_schema_1_0
        )
        self.assertIsInstance(result, utils.RuleProcessor)

    def test_create_attribute_mapping_rules_processor_schema1_0(self):
        result = utils.create_attribute_mapping_rules_processor(
            self.attribute_mapping_schema_1_0
        )
        self.assertIsInstance(result, utils.RuleProcessor)

    def test_create_attribute_mapping_rules_processor_schema2_0(self):
        result = utils.create_attribute_mapping_rules_processor(
            self.attribute_mapping_schema_2_0
        )
        self.assertIsInstance(result, utils.RuleProcessorToHonorDomainOption)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import unittest

from testtools import matchers
from testtools import testcase

from keystone.tests.unit import utils


class TestWipDecorator(testcase.TestCase):
    def test_raises_SkipError_when_broken_test_fails(self):
        @utils.wip('waiting on bug #000000')
        def test():
            raise Exception('i expected a failure - this is a WIP')

        e = self.assertRaises(unittest.SkipTest, test)
        self.assertThat(str(e), matchers.Contains('#000000'))

    def test_raises_AssertionError_when_test_passes(self):
        @utils.wip('waiting on bug #000000')
        def test():
            pass  # literally

        e = self.assertRaises(AssertionError, test)
        self.assertThat(str(e), matchers.Contains('#000000'))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from typing import Any

from keystone.api.validation import response_types

# Individual properties of the `Credential`
_credential_properties = {
    "blob": {
        "type": "string",
        "description": "The credential itself, as a serialized blob.",
    },
    "project_id": {
        "type": ["string", "null"],
        "description": "The ID for the project. Mandatory for `EC2` type.",
    },
    "type": {
        "type": "string",
        "description": (
            "The credential type, such as ec2 or cert. The implementation"
            " determines the list of supported types."
        ),
    },
    "user_id": {
        "type": "string",
        "format": "uuid",
        "description": "The ID of the user who owns the credential.",
    },
}

# Common schema of the `Credential` resource
credential_schema: dict[str, Any] = {
    "type": "object",
    "description": "A credential object.",
    "properties": {
        "id": {
            "type": "string",
            "readOnly": True,
            "description": "The UUID for the credential.",
        },
        "links": response_types.resource_links,
        **_credential_properties,
    },
    "additionalProperties": True,
}

# Query parameters of the `/credentials` API
index_request_query: dict[str, Any] = {
    "type": "object",
    "properties": {
        "user_id": {
            "type": "string",
            "format": "uuid",
            "description": "Filters the response by a user ID.",
        },
        "type": {
            "type": "string",
            "description": (
                "The credential type, such as ec2 or cert. The implementation"
                " determines the list of supported types."
            ),
        },
    },
}

# Response of the `/credentials` API
index_response_body: dict[str, Any] = {
    "type": "object",
    "properties": {
        "credentials": {
            "type": "array",
            "items": credential_schema,
            "description": "A list of credential objects.",
        },
        "links": response_types.links,
        "truncated": response_types.truncated,
    },
    "additionalProperties": False,
}

# Response of the diverse `/credentials` APIs returning single
# credential
credential_response_body: dict[str, Any] = {
    "type": "object",
    "description": "A credential object.",
    "properties": {"credential": credential_schema},
    "additionalProperties": False,
}

# Request body of the `POST /credentials` operation
create_request_body: dict[str, Any] = {
    "type": "object",
    "description": "A credential object.",
    "properties": {
        "credential": {
            "type": "object",
            "properties": {
                "id": {
                    "type": "string",
                    "description": "The UUID for the credential.",
                },
                **_credential_properties,
            },
            "additionalProperties": True,
            "required": ["blob", "type", "user_id"],
            # Generating client code for conditionally optional parameters is
            # hard. Use if-then-else for validation keeping the schema itself
            # static.
            "if": {"properties": {"type": {"const": "ec2"}}},
            "then": {
                "title": "ec2 credential requires project_id",
                "required": ["blob", "type", "user_id", "project_id"],
            },
        }
    },
    "required": ["credential"],
}

# Request body of the `PATCH /credentials/{credential_id}` operation
update_request_body: dict[str, Any] = {
    "type": "object",
    "description": "A credential object.",
    "properties": {
        "credential": {
            "type": "object",
            "properties": _credential_properties,
            "additionalProperties": True,
            "minProperties": 1,
        }
    },
    "required": ["credential"],
}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from keystone.common import utils
from keystone import exception
from keystone.tests import unit


class FormatUrlTests(unit.BaseTestCase):
    def test_successful_formatting(self):
        url_template = (
            'http://server:9090/$(tenant_id)s/$(user_id)s/$(project_id)s'
        )
        project_id = uuid.uuid4().hex
        values = {'tenant_id': 'A', 'user_id': 'B', 'project_id': project_id}
        actual_url = utils.format_url(url_template, values)

        expected_url = f'http://server:9090/A/B/{project_id}'
        self.assertEqual(expected_url, actual_url)

    def test_raises_malformed_on_missing_key(self):
        self.assertRaises(
            exception.MalformedEndpoint,
            utils.format_url,
            "http://server:9090/$(tenant_id)s",
            {},
        )

    def test_raises_malformed_on_wrong_type(self):
        self.assertRaises(
            exception.MalformedEndpoint,
            utils.format_url,
            "http://server:9090/$(tenant_id)d",
            {"tenant_id": 'A'},
        )

    def test_raises_malformed_on_incomplete_format(self):
        self.assertRaises(
            exception.MalformedEndpoint,
            utils.format_url,
            "http://server:9090/$(tenant_id)",
            {"tenant_id": 'A'},
        )

    def test_formatting_a_non_string(self):
        def _test(url_template):
            self.assertRaises(
                exception.MalformedEndpoint, utils.format_url, url_template, {}
            )

        _test(None)
        _test(object())

    def test_substitution_with_key_not_allowed(self):
        # If the url template contains a substitution that's not in the allowed
        # list then MalformedEndpoint is raised.
        # For example, admin_token isn't allowed.
        url_template = (
            'http://server:9090/$(project_id)s/$(user_id)s/$(admin_token)s'
        )
        values = {'user_id': 'B', 'admin_token': 'C'}
        self.assertRaises(
            exception.MalformedEndpoint, utils.format_url, url_template, values
        )

    def test_substitution_with_allowed_tenant_keyerror(self):
        # No value of 'tenant_id' is passed into url_template.
        # mod: format_url will return None instead of raising
        # "MalformedEndpoint" exception.
        # This is intentional behavior since we don't want to skip
        # all the later endpoints once there is an URL of endpoint
        # trying to replace 'tenant_id' with None.
        url_template = 'http://server:9090/$(tenant_id)s/$(user_id)s'
        values = {'user_id': 'B'}
        self.assertIsNone(
            utils.format_url(
                url_template, values, silent_keyerror_failures=['tenant_id']
            )
        )

    def test_substitution_with_allowed_project_keyerror(self):
        # No value of 'project_id' is passed into url_template.
        # mod: format_url will return None instead of raising
        # "MalformedEndpoint" exception.
        # This is intentional behavior since we don't want to skip
        # all the later endpoints once there is an URL of endpoint
        # trying to replace 'project_id' with None.
        url_template = 'http://server:9090/$(project_id)s/$(user_id)s'
        values = {'user_id': 'B'}
        self.assertIsNone(
            utils.format_url(
                url_template, values, silent_keyerror_failures=['project_id']
            )
        )



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import fixtures

from keystone import auth
import keystone.server


class BackendLoader(fixtures.Fixture):
    """Initialize each manager and assigns them to an attribute."""

    def __init__(self, testcase):
        super().__init__()
        self._testcase = testcase

    def setUp(self):
        super().setUp()

        self.clear_auth_plugin_registry()
        drivers, _unused = keystone.server.setup_backends()

        for manager_name, manager in drivers.items():
            setattr(self._testcase, manager_name, manager)

        self.addCleanup(self._testcase.cleanup_instance(*list(drivers.keys())))

        del self._testcase  # break circular reference

    def clear_auth_plugin_registry(self):
        auth.core.AUTH_METHODS.clear()
        auth.core.AUTH_PLUGINS_LOADED = False



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

from keystone.conf import utils

driver = cfg.StrOpt(
    'driver',
    default='sql',
    help=utils.fmt(
        """
Entry point for the assignment backend driver (where role assignments are
stored) in the `keystone.assignment` namespace. Only a SQL driver is supplied
by keystone itself. Unless you are writing proprietary drivers for keystone,
you do not need to set this option.
"""
    ),
)

prohibited_implied_role = cfg.ListOpt(
    'prohibited_implied_role',
    default=['admin'],
    help=utils.fmt(
        """
A list of role names which are prohibited from being an implied role.
"""
    ),
)


GROUP_NAME = __name__.split('.')[-1]
ALL_OPTS = [driver, prohibited_implied_role]


def register_opts(conf):
    conf.register_opts(ALL_OPTS, group=GROUP_NAME)


def list_opts():
    return {GROUP_NAME: ALL_OPTS}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any

from keystone.api.validation import parameter_types
from keystone.api.validation import response_types
from keystone.common import validation

_region_properties = {
    'description': validation.nullable(parameter_types.description),
    # NOTE(lbragstad): Regions use ID differently. The user can specify the ID
    # or it will be generated automatically.
    'id': {'type': 'string'},
    'parent_region_id': {'type': ['string', 'null']},
}

region_create = {
    'type': 'object',
    'properties': _region_properties,
    'additionalProperties': True,
    # NOTE(lbragstad): No parameters are required for creating regions.
}

region_update = {
    'type': 'object',
    'properties': _region_properties,
    'minProperties': 1,
    'additionalProperties': True,
}

# Schema for Service v3
# Individual properties of 'Service'
_service_properties = {
    "enabled": {
        "type": "boolean",
        "description": (
            "Defines whether the service and its endpoints appear in the "
            "service catalog - false. The service and its endpoints do "
            "not appear in the service catalog - true."
        ),
    },
    "type": {
        "type": "string",
        "description": (
            "The service type, which describes the API implemented by the "
            "service. Value is compute, ec2, identity, image, network, "
            "or volume."
        ),
        "minLength": 1,
        "maxLength": 255,
    },
}

_service_name_properties = {
    "name": {
        "type": "string",
        "description": "The service name.",
        "minLength": 1,
        "maxLength": 255,
    }
}

# Common schema of `Service` resource
service_schema: dict[str, Any] = {
    "type": "object",
    "description": "A service object",
    "properties": {
        "id": {
            "type": "string",
            "readOnly": True,
            "description": (
                "The UUID of the service to which the endpoint belongs."
            ),
        },
        "name": {
            "type": "string",
            "description": "The service name.",
            "minLength": 0,
            "maxLength": 255,
        },
        "links": response_types.resource_links,
        **_service_properties,
    },
    "additionalProperties": True,
}

# Query parameters of the `/services` API
service_index_request_query: dict[str, Any] = {
    "type": "object",
    "properties": {},
    "additionalProperties": True,
}

# Response of the `/services` API
service_index_response_body: dict[str, Any] = {
    "type": "object",
    "properties": {
        "services": {
            "type": "array",
            "items": service_schema,
            "description": "A list of service object.",
        },
        "links": response_types.links,
        "truncated": response_types.truncated,
    },
    "additionalProperties": False,
}

# Response of the `/services` API returning a single service
service_response_body: dict[str, Any] = {
    "type": "object",
    "description": "A service object.",
    "properties": {"service": service_schema},
    "additionalProperties": False,
}

# Request body of the `POST /services` operation
service_create_request_body: dict[str, Any] = {
    "type": "object",
    "description": "A service object.",
    "properties": {
        "service": {
            "type": "object",
            "properties": {**_service_properties, **_service_name_properties},
            "additionalProperties": True,
            "required": ["type"],
        }
    },
}

# Request body of the `PATCH /services/{service_id}` operation
service_update_request_body: dict[str, Any] = {
    "type": "object",
    "description": "A service object.",
    "properties": {
        "service": {
            "type": "object",
            "properties": {**_service_properties, **_service_name_properties},
            "additionalProperties": True,
            "minProperties": 1,
        }
    },
    "required": ["service"],
}

# Individual properties of 'Endpoint'
_endpoint_properties = {
    "enabled": {
        "type": "boolean",
        "description": (
            "Indicates whether the endpoint appears in the service "
            "catalog -false. The endpoint does not appear in the service "
            "catalog. -true. The endpoint appears in the service catalog."
        ),
    },
    "interface": {
        "type": "string",
        "enum": ["admin", "internal", "public"],
        "description": (
            "The interface type, which describes the visibility of the "
            "endpoint. Value is: -public. Visible by end users on a "
            "publicly available network interface. -internal. Visible "
            "by end users on an unmetered internal network interface. -admin. "
            "Visible by administrative users on a secure network interface."
        ),
    },
    "region_id": {
        "type": ["string", "null"],
        "description": (
            "(Since v3.2) The ID of the region that contains the "
            "service endpoint."
        ),
        "x-openstack": {"min-ver": 3.2},
    },
    "region": {
        "type": ["string", "null"],
        "description": (
            "(Deprecated in v3.2) The geographic location of "
            "the service endpoint."
        ),
        "x-openstack": {"max-ver": 3.2},
    },
    "service_id": {
        "type": "string",
        "description": "The UUID of the service to which the endpoint belongs",
    },
    "url": {
        "type": "string",
        "description": "The endpoint URL.",
        "minLength": 0,
        "maxLength": 225,
        "pattern": "^[a-zA-Z0-9+.-]+:.+",
    },
    "name": {"type": "string", "description": "The name of the endpoint."},
    "description": {
        "type": ["string", "null"],
        "description": "A description of the endpoint.",
    },
}

# Common schema of `Endpoint` resource
endpoint_schema: dict[str, Any] = {
    "type": "object",
    "description": "An endpoint object",
    "properties": {
        "id": {
            "type": "string",
            "readOnly": True,
            "description": "The endpoint ID",
        },
        "links": response_types.resource_links,
        **_endpoint_properties,
    },
    "additionalProperties": False,
}

# Query parameters of the `/endpoints` API
endpoint_index_request_query: dict[str, Any] = {
    "type": "object",
    "properties": {
        "interface": {
            "type": "string",
            "enum": ["admin", "internal", "public"],
            "description": (
                "The interface type, which describes the visibility of the "
                "endpoint. Value is: -public. Visible by end users on a "
                "publicly available network interface. -internal. Visible "
                "by end users on an unmetered internal network interface."
                "-admin. Visible by administrative users on a secure "
                "network interface."
            ),
        },
        "region_id": {
            "type": ["string", "null"],
            "description": (
                "(Since v3.2) The ID of the region that contains the "
                "service endpoint."
            ),
            "x-openstack": {"min-ver": 3.2},
        },
        "service_id": {
            "type": "string",
            "description": "The UUID of the service to which the "
            "endpoint belongs",
        },
    },
    "additionalProperties": False,
}

# Response of the `/endpoints` API
endpoint_index_response_body: dict[str, Any] = {
    "type": "object",
    "properties": {
        "endpoints": {
            "type": "array",
            "items": endpoint_schema,
            "description": "A list of endpoint objects.",
        },
        "links": response_types.links,
        "truncated": response_types.truncated,
    },
    "additionalProperties": False,
}

endpoint_request_query: dict[str, Any] = {
    "type": "object",
    "properties": {},
    "additionalProperties": False,
}

# Response of the `/endpoints` API returning a single endpoint
endpoint_response_body: dict[str, Any] = {
    "type": "object",
    "description": "An endpoint object",
    "properties": {"endpoint": endpoint_schema},
    "additionalProperties": False,
}

# Request body of the `POST /endpoints` operation
endpoint_create_request_body: dict[str, Any] = {
    "type": "object",
    "description": "An endpoint object",
    "properties": {
        "endpoint": {
            "type": "object",
            "properties": {
                "id": {"type": "string", "description": "The endpoint ID."},
                **_endpoint_properties,
            },
            "required": ["interface", "service_id", "url"],
            "additionalProperties": True,
        }
    },
    "additionalProperties": False,
}

# Request body of the `PATCH /endpoints/{endpoint_id}`
endpoint_update_request_body: dict[str, Any] = {
    "type": "object",
    "description": "An endpoint object",
    "properties": {
        "endpoint": {
            "type": "object",
            "properties": _endpoint_properties,
            "additionalProperties": True,
            "minProperties": 1,
        }
    },
    "required": ["endpoint"],
    "additionalProperties": False,
}

# Individual properties of 'Endpoint Group'
_endpoint_group_properties = {
    "description": {
        "type": ["string", "null"],
        "description": "The endpoint group description.",
    },
    "filters": {
        "type": "object",
        "description": (
            "Describes the filtering performed by the endpoint group. "
            "The filter used must be an endpoint property, such as "
            "interface, service_id, region, and enabled. Note that "
            "if using interface as a filter, the only available values "
            "are public, internal, and admin."
        ),
        "properties": {
            "interface": {
                "type": "string",
                "enum": ["admin", "internal", "public"],
                "description": (
                    "The interface type, which describes the visibility of "
                    "the endpoint. Value is: -public. Visible by end users "
                    "on a publicly available network interface. -internal. "
                    "Visible by end users on an unmetered internal network "
                    "interface. -admin. Visible by administrative users on "
                    "a secure network interface."
                ),
            },
            "service_id": {
                "type": "string",
                "description": "The UUID of the service to which the "
                "endpoint belongs",
            },
            "region_id": {
                "type": ["string", "null"],
                "description": (
                    "(Since v3.2) The ID of the region that contains the "
                    "service endpoint."
                ),
                "x-openstack": {"min-ver": 3.2},
            },
            "enabled": {
                "type": "boolean",
                "description": (
                    "Indicates whether the endpoint appears in the service "
                    "catalog -false. The endpoint does not appear in the "
                    "service catalog. -true. The endpoint appears in the "
                    "service catalog."
                ),
            },
        },
    },
    "name": {
        **parameter_types.name,
        "description": "The name of the endpoint group.",
    },
}

# Common schema of `Endpoint Group` resource
endpoint_group_schema: dict[str, Any] = {
    "type": "object",
    "description": "An endpoint group object.",
    "properties": {
        "id": {
            "type": "string",
            "readOnly": True,
            "description": "The endpoint group ID",
        },
        "links": response_types.resource_links,
        **_endpoint_group_properties,
    },
    "additionalProperties": False,
}

# Query parameters of the `/endpoint_groups` API
endpoint_group_index_request_query: dict[str, Any] = {
    "type": "object",
    "properties": {
        "name": {
            **parameter_types.name,
            "description": "The name of the endpoint group.",
        }
    },
    "additionalProperties": False,
}

# Response of the `/endpoint_groups` API
endpoint_group_index_response_body: dict[str, Any] = {
    "type": "object",
    "properties": {
        "endpoint_groups": {
            "type": "array",
            "items": endpoint_group_schema,
            "description": "A list of endpoint group objects",
        },
        "links": response_types.links,
        "truncated": response_types.truncated,
    },
    "additionalProperties": False,
}

# Response of the `/endpoint_groups/{endpoint_group_id}` API
# returning a single endpoint group
endpoint_group_response_body: dict[str, Any] = {
    "type": "object",
    "description": "An endpoint group object",
    "properties": {"endpoint_group": endpoint_group_schema},
    "additionalProperties": False,
}

# Request body of the `POST /endpoint_groups` operation
endpoint_group_create_request_body: dict[str, Any] = {
    "type": "object",
    "description": "An endpoint group object",
    "properties": {
        "endpoint_group": {
            "type": "object",
            "properties": _endpoint_group_properties,
            "required": ["name", "filters"],
            "additionalProperties": False,
        }
    },
    "additionalProperties": False,
}

# Request body of the `PATCH /endpoint_groups/{endpoint_group_id}` operation
endpoint_group_update_request_body: dict[str, Any] = {
    "type": "object",
    "description": "An endpoint group object",
    "properties": {
        "endpoint_group": {
            "type": "object",
            "properties": _endpoint_group_properties,
            "minProperties": 1,
            "additionalProperties": False,
        }
    },
    "required": ["endpoint_group"],
    "additionalProperties": False,
}



#!/usr/bin/env python3
# Copyright 2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import sys

from alembic import command as alembic_command
from alembic import script as alembic_script
from alembic import util as alembic_util
from oslo_config import cfg
from oslo_log import log
import pbr.version

from keystone.common import sql
from keystone.common.sql import upgrades
import keystone.conf
from keystone.i18n import _

CONF = keystone.conf.CONF
LOG = log.getLogger(__name__)


def import_sql_modules():
    # We need to import all of these so the tables are registered. It would be
    # easier if these were all in a central location :(
    import keystone.application_credential.backends.sql  # noqa: F401
    import keystone.assignment.backends.sql  # noqa: F401
    import keystone.assignment.role_backends.sql_model  # noqa: F401
    import keystone.catalog.backends.sql  # noqa: F401
    import keystone.credential.backends.sql  # noqa: F401
    import keystone.endpoint_policy.backends.sql  # noqa: F401
    import keystone.federation.backends.sql  # noqa: F401
    import keystone.identity.backends.sql_model  # noqa: F401
    import keystone.identity.mapping_backends.sql  # noqa: F401
    import keystone.limit.backends.sql  # noqa: F401
    import keystone.oauth1.backends.sql  # noqa: F401
    import keystone.policy.backends.sql  # noqa: F401
    import keystone.resource.backends.sql_model  # noqa: F401
    import keystone.resource.config_backends.sql  # noqa: F401
    import keystone.revoke.backends.sql  # noqa: F401
    import keystone.trust.backends.sql  # noqa: F401


def do_alembic_command(config, cmd, revision=None, **kwargs):
    args = []
    if revision:
        args.append(revision)

    try:
        getattr(alembic_command, cmd)(config, *args, **kwargs)
    except alembic_util.CommandError as e:
        alembic_util.err(str(e))


def do_generic_show(config, cmd):
    kwargs = {'verbose': CONF.command.verbose}
    do_alembic_command(config, cmd, **kwargs)


def do_validate(config, cmd):
    do_alembic_command(config, 'branches')
    # TODO(stephenfin): Implement these
    # validate_revisions(config)
    # TODO(stephenfin): Implement these
    # validate_head_files(config)


def _find_milestone_revisions(config, milestone, branch=None):
    """Return the revision(s) for a given milestone."""
    script = alembic_script.ScriptDirectory.from_config(config)
    return [
        (m.revision, label)
        for m in _get_revisions(script)
        for label in (m.branch_labels or [None])
        if milestone in getattr(m.module, 'keystone_milestone', [])
        and (branch is None or branch in m.branch_labels)
    ]


def _get_revisions(script):
    return list(script.walk_revisions(base='base', head='heads'))


def do_upgrade(config, cmd):
    branch = None

    if (CONF.command.revision or CONF.command.delta) and (
        CONF.command.expand or CONF.command.contract
    ):
        msg = _('Phase upgrade options do not accept revision specification')
        raise SystemExit(msg)

    if CONF.command.expand:
        branch = upgrades.EXPAND_BRANCH
        revision = f'{upgrades.EXPAND_BRANCH}@head'
    elif CONF.command.contract:
        branch = upgrades.CONTRACT_BRANCH
        revision = f'{upgrades.CONTRACT_BRANCH}@head'
    elif not CONF.command.revision and not CONF.command.delta:
        msg = _('You must provide a revision or relative delta')
        raise SystemExit(msg)
    else:
        revision = CONF.command.revision or ''
        if '-' in revision:
            msg = _('Negative relative revision (downgrade) not supported')
            raise SystemExit(msg)

        delta = CONF.command.delta
        if delta:
            if '+' in revision:
                msg = _('Use either --delta or relative revision, not both')
                raise SystemExit(msg)
            if delta < 0:
                msg = _('Negative delta (downgrade) not supported')
                raise SystemExit(msg)
            revision = '%s+%d' % (revision, delta)

        # leave branchless 'head' revision request backward compatible by
        # applying all heads in all available branches.
        if revision == 'head':
            revision = 'heads'

    if revision in upgrades.MILESTONES:
        expand_revisions = _find_milestone_revisions(
            config, revision, upgrades.EXPAND_BRANCH
        )
        contract_revisions = _find_milestone_revisions(
            config, revision, upgrades.CONTRACT_BRANCH
        )
        # Expand revisions must be run before contract revisions
        revisions = expand_revisions + contract_revisions
    else:
        revisions = [(revision, branch)]

    for revision, branch in revisions:
        # if not CONF.command.sql:
        #     run_sanity_checks(config, revision)
        do_alembic_command(
            config, cmd, revision=revision, sql=CONF.command.sql
        )


def do_revision(config, cmd):
    kwargs = {
        'message': CONF.command.message,
        'autogenerate': CONF.command.autogenerate,
        'sql': CONF.command.sql,
    }
    branches = []
    if CONF.command.expand:
        kwargs['head'] = 'expand@head'
        branches.append(upgrades.EXPAND_BRANCH)
    elif CONF.command.contract:
        kwargs['head'] = 'contract@head'
        branches.append(upgrades.CONTRACT_BRANCH)
    else:
        branches = upgrades.MIGRATION_BRANCHES

    if not CONF.command.autogenerate:
        for branch in branches:
            args = copy.copy(kwargs)
            version_path = upgrades.get_version_branch_path(
                release=upgrades.CURRENT_RELEASE, branch=branch
            )
            upgrades.check_bootstrap_new_branch(branch, version_path, args)
            do_alembic_command(config, cmd, **args)
    else:  # CONF.command.autogenerate
        # autogeneration code will take care of enforcing proper directories
        do_alembic_command(config, cmd, **kwargs)

    # TODO(stephenfin): Implement these
    # update_head_files(config)


def add_branch_options(parser):
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--expand', action='store_true')
    group.add_argument('--contract', action='store_true')
    return group


def add_alembic_subparser(sub, cmd):
    return sub.add_parser(cmd, help=getattr(alembic_command, cmd).__doc__)


def add_command_parsers(subparsers):
    for name in ['current', 'history', 'branches', 'heads']:
        parser = add_alembic_subparser(subparsers, name)
        parser.set_defaults(func=do_generic_show)
        parser.add_argument(
            '--verbose',
            action='store_true',
            help='Display more verbose output for the specified command',
        )

    parser = add_alembic_subparser(subparsers, 'upgrade')
    parser.add_argument('--delta', type=int)
    parser.add_argument('--sql', action='store_true')
    parser.add_argument('revision', nargs='?')
    add_branch_options(parser)
    parser.set_defaults(func=do_upgrade)

    parser = subparsers.add_parser(
        'validate',
        help=alembic_command.branches.__doc__ + ' and validate head file',
    )
    parser.set_defaults(func=do_validate)

    parser = add_alembic_subparser(subparsers, 'revision')
    parser.add_argument('-m', '--message')
    parser.add_argument('--sql', action='store_true')
    group = add_branch_options(parser)
    group.add_argument('--autogenerate', action='store_true')
    parser.set_defaults(func=do_revision)


command_opt = cfg.SubCommandOpt(
    'command',
    title='Command',
    help=_('Available commands'),
    handler=add_command_parsers,
)


def main(argv):
    CONF.register_cli_opt(command_opt)

    keystone.conf.configure()
    sql.initialize()
    keystone.conf.set_default_for_default_log_levels()

    user_supplied_config_file = False
    if argv:
        for argument in argv:
            if argument == '--config-file':
                user_supplied_config_file = True

    CONF(
        project='keystone',
        version=pbr.version.VersionInfo('keystone').version_string(),
    )

    if not CONF.default_config_files and not user_supplied_config_file:
        LOG.warning('Config file not found, using default configs.')

    import_sql_modules()

    config = upgrades.get_alembic_config()

    return bool(CONF.command.func(config, CONF.command.name))


if __name__ == '__main__':
    main(sys.argv)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from oslo_log import log

from keystone.common import provider_api
from keystone.common.validation import validators
import keystone.conf
from keystone import exception
from keystone.identity.mapping_backends import mapping
from keystone import notifications
from keystone.server import backends

CONF = keystone.conf.CONF
LOG = log.getLogger(__name__)
PROVIDERS = provider_api.ProviderAPIs


class Identity:
    def __init__(self):
        backends.load_backends()

        self.user_id = None
        self.user_name = None
        self.user_password = None

        self.project_id = None
        self.project_name = None

        self.default_domain_id = CONF.identity.default_domain_id

    def project_setup(self):
        try:
            project_id = self.project_id
            if project_id is None:
                project_id = uuid.uuid4().hex
            project = {
                'enabled': True,
                'id': project_id,
                'domain_id': self.default_domain_id,
                'description': 'Bootstrap project for initializing the cloud.',
                'name': self.project_name,
            }
            PROVIDERS.resource_api.create_project(project_id, project)
            LOG.info('Created project %s', self.project_name)
        except exception.Conflict:
            LOG.info(
                'Project %s already exists, skipping creation.',
                self.project_name,
            )
            project = PROVIDERS.resource_api.get_project_by_name(
                self.project_name, self.default_domain_id
            )

        self.project_id = project['id']

    def _create_user(self, user_ref, initiator=None):
        _self = PROVIDERS.identity_api.create_user.__self__
        user = user_ref.copy()
        if 'password' in user:
            validators.validate_password(user['password'])
        user['name'] = user['name'].strip()
        user.setdefault('enabled', True)
        domain_id = user['domain_id']
        PROVIDERS.resource_api.get_domain(domain_id)

        _self._assert_default_project_id_is_not_domain(
            user_ref.get('default_project_id')
        )

        # For creating a user, the domain is in the object itself
        domain_id = user_ref['domain_id']
        driver = _self._select_identity_driver(domain_id)
        user = _self._clear_domain_id_if_domain_unaware(driver, user)
        # Generate a local ID - in the future this might become a function of
        # the underlying driver so that it could conform to rules set down by
        # that particular driver type.
        user['id'] = self.user_id
        ref = _self._create_user_with_federated_objects(user, driver)
        notifications.Audit.created(_self._USER, user['id'], initiator)
        return _self._set_domain_id_and_mapping(
            ref, domain_id, driver, mapping.EntityType.USER
        )

    def user_setup(self):
        # NOTE(morganfainberg): Do not create the user if it already exists.
        try:
            user = PROVIDERS.identity_api.get_user_by_name(
                self.user_name, self.default_domain_id
            )
            LOG.info(
                'User %s already exists, skipping creation.', self.user_name
            )

            if self.user_id is not None and user['id'] != self.user_id:
                msg = (
                    f'user `{self.user_name}` already exists '
                    f'with `{self.user_id}`'
                )
                raise exception.Conflict(type='user_id', details=msg)

            # If the user is not enabled, re-enable them. This also helps
            # provide some useful logging output later.
            update = {}
            enabled = user['enabled']
            if not enabled:
                update['enabled'] = True

            try:
                PROVIDERS.identity_api.driver.authenticate(
                    user['id'], self.user_password
                )
            except AssertionError:
                # This means that authentication failed and that we need to
                # update the user's password. This is going to persist a
                # revocation event that will make all previous tokens for the
                # user invalid, which is OK because it falls within the scope
                # of revocation. If a password changes, we shouldn't be able to
                # use tokens obtained with an old password.
                update['password'] = self.user_password

            # Only make a call to update the user if the password has changed
            # or the user was previously disabled. This allows bootstrap to act
            # as a recovery tool, without having to create a new user.
            if update:
                user = PROVIDERS.identity_api.update_user(user['id'], update)
                LOG.info('Reset password for user %s.', self.user_name)
                if not enabled and user['enabled']:
                    # Although we always try to enable the user, this log
                    # message only makes sense if we know that the user was
                    # previously disabled.
                    LOG.info('Enabled user %s.', self.user_name)
        except exception.UserNotFound:
            user = self._create_user(
                user_ref={
                    'name': self.user_name,
                    'enabled': True,
                    'domain_id': self.default_domain_id,
                    'password': self.user_password,
                }
            )
            LOG.info('Created user %s', self.user_name)

        self.user_id = user['id']



# Copyright 2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import ldappool

from keystone.common import provider_api
import keystone.conf
from keystone.identity.backends import ldap
from keystone.identity.backends.ldap import common as ldap_common
from keystone.tests import unit
from keystone.tests.unit import fakeldap
from keystone.tests.unit import test_backend_ldap_pool
from keystone.tests.unit import test_ldap_livetest

CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs


class LiveLDAPPoolIdentity(
    test_backend_ldap_pool.LdapPoolCommonTestMixin,
    test_ldap_livetest.LiveLDAPIdentity,
):
    """Executes existing LDAP live test with pooled LDAP handler.

    Also executes common pool specific tests via Mixin class.

    """

    def setUp(self):
        super().setUp()
        self.addCleanup(self.cleanup_pools)
        # storing to local variable to avoid long references
        self.conn_pools = ldap_common.PooledLDAPHandler.connection_pools

    def config_files(self):
        config_files = super().config_files()
        config_files.append(unit.dirs.tests_conf('backend_pool_liveldap.conf'))
        return config_files

    def test_assert_connector_used_not_fake_ldap_pool(self):
        handler = ldap_common._get_connection(CONF.ldap.url, use_pool=True)
        self.assertNotEqual(
            type(handler.Connector), type(fakeldap.FakeLdapPool)
        )
        self.assertEqual(
            type(ldappool.StateConnector), type(handler.Connector)
        )

    def test_async_search_and_result3(self):
        self.config_fixture.config(group='ldap', page_size=1)
        self.test_user_enable_attribute_mask()

    def test_pool_size_expands_correctly(self):
        who = CONF.ldap.user
        cred = CONF.ldap.password
        # get related connection manager instance
        ldappool_cm = self.conn_pools[CONF.ldap.url]

        def _get_conn():
            return ldappool_cm.connection(who, cred)

        with _get_conn() as c1:  # 1
            self.assertEqual(1, len(ldappool_cm))
            self.assertTrue(c1.connected)
            self.assertTrue(c1.active)
            with _get_conn() as c2:  # conn2
                self.assertEqual(2, len(ldappool_cm))
                self.assertTrue(c2.connected)
                self.assertTrue(c2.active)

            self.assertEqual(2, len(ldappool_cm))
            # c2 went out of context, its connected but not active
            self.assertTrue(c2.connected)
            self.assertFalse(c2.active)
            with _get_conn() as c3:  # conn3
                self.assertEqual(2, len(ldappool_cm))
                self.assertTrue(c3.connected)
                self.assertTrue(c3.active)
                self.assertIs(c3, c2)  # same connection is reused
                self.assertTrue(c2.active)
                with _get_conn() as c4:  # conn4
                    self.assertEqual(3, len(ldappool_cm))
                    self.assertTrue(c4.connected)
                    self.assertTrue(c4.active)

    def test_password_change_with_auth_pool_disabled(self):
        self.config_fixture.config(group='ldap', use_auth_pool=False)
        old_password = self.user_sna['password']

        self.test_password_change_with_pool()

        self.assertRaises(
            AssertionError,
            PROVIDERS.identity_api.authenticate,
            context={},
            user_id=self.user_sna['id'],
            password=old_password,
        )

    def _create_user_and_authenticate(self, password):
        user = unit.create_user(
            PROVIDERS.identity_api,
            CONF.identity.default_domain_id,
            password=password,
        )

        with self.make_request():
            PROVIDERS.identity_api.authenticate(
                user_id=user['id'], password=password
            )

        return PROVIDERS.identity_api.get_user(user['id'])

    def _get_auth_conn_pool_cm(self):
        pool_url = (
            ldap_common.PooledLDAPHandler.auth_pool_prefix + CONF.ldap.url
        )
        return self.conn_pools[pool_url]

    def _do_password_change_for_one_user(self, password, new_password):
        self.config_fixture.config(group='ldap', use_auth_pool=True)
        self.cleanup_pools()
        self.load_backends()

        user1 = self._create_user_and_authenticate(password)
        auth_cm = self._get_auth_conn_pool_cm()
        self.assertEqual(1, len(auth_cm))
        user2 = self._create_user_and_authenticate(password)
        self.assertEqual(1, len(auth_cm))
        user3 = self._create_user_and_authenticate(password)
        self.assertEqual(1, len(auth_cm))
        user4 = self._create_user_and_authenticate(password)
        self.assertEqual(1, len(auth_cm))
        user5 = self._create_user_and_authenticate(password)
        self.assertEqual(1, len(auth_cm))

        # connection pool size remains 1 even for different user ldap bind
        # as there is only one active connection at a time

        user_api = ldap.UserApi(CONF)
        u1_dn = user_api._id_to_dn_string(user1['id'])
        u2_dn = user_api._id_to_dn_string(user2['id'])
        u3_dn = user_api._id_to_dn_string(user3['id'])
        u4_dn = user_api._id_to_dn_string(user4['id'])
        u5_dn = user_api._id_to_dn_string(user5['id'])

        # now create multiple active connections for end user auth case which
        # will force to keep them in pool. After that, modify one of user
        # password. Need to make sure that user connection is in middle
        # of pool list.
        auth_cm = self._get_auth_conn_pool_cm()
        with auth_cm.connection(u1_dn, password) as _:
            with auth_cm.connection(u2_dn, password) as _:
                with auth_cm.connection(u3_dn, password) as _:
                    with auth_cm.connection(u4_dn, password) as _:
                        with auth_cm.connection(u5_dn, password) as _:
                            self.assertEqual(5, len(auth_cm))
                            _.unbind_s()

        user3['password'] = new_password
        PROVIDERS.identity_api.update_user(user3['id'], user3)

        return user3

    def test_password_change_with_auth_pool_enabled_long_lifetime(self):
        self.config_fixture.config(
            group='ldap', auth_pool_connection_lifetime=600
        )
        old_password = 'my_password'
        new_password = 'new_password'
        user = self._do_password_change_for_one_user(
            old_password, new_password
        )
        user.pop('password')

        # with long connection lifetime auth_pool can bind to old password
        # successfully which is not desired if password change is frequent
        # use case in a deployment.
        # This can happen in multiple concurrent connections case only.
        with self.make_request():
            user_ref = PROVIDERS.identity_api.authenticate(
                user_id=user['id'], password=old_password
            )

        self.assertDictEqual(user, user_ref)

    def test_password_change_with_auth_pool_enabled_no_lifetime(self):
        self.config_fixture.config(
            group='ldap', auth_pool_connection_lifetime=0
        )

        old_password = 'my_password'
        new_password = 'new_password'
        user = self._do_password_change_for_one_user(
            old_password, new_password
        )
        # now as connection lifetime is zero, so authentication
        # with old password will always fail.
        self.assertRaises(
            AssertionError,
            PROVIDERS.identity_api.authenticate,
            context={},
            user_id=user['id'],
            password=old_password,
        )



# Copyright 2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import uuid

from keystone.common import provider_api
from keystone.common.resource_options import options as ro_opt
from keystone import exception
from keystone.tests import unit
from keystone.tests.unit import default_fixtures

PROVIDERS = provider_api.ProviderAPIs


class RoleTests:
    def test_get_role_returns_not_found(self):
        self.assertRaises(
            exception.RoleNotFound,
            PROVIDERS.role_api.get_role,
            uuid.uuid4().hex,
        )

    def test_get_unique_role_by_name_returns_not_found(self):
        self.assertRaises(
            exception.RoleNotFound,
            PROVIDERS.role_api.get_unique_role_by_name,
            uuid.uuid4().hex,
        )

    def test_create_duplicate_role_name_fails(self):
        role_id = uuid.uuid4().hex
        role = unit.new_role_ref(id=role_id, name='fake1name')
        PROVIDERS.role_api.create_role(role_id, role)
        new_role_id = uuid.uuid4().hex
        role['id'] = new_role_id
        self.assertRaises(
            exception.Conflict,
            PROVIDERS.role_api.create_role,
            new_role_id,
            role,
        )

    def test_rename_duplicate_role_name_fails(self):
        role_id1 = uuid.uuid4().hex
        role_id2 = uuid.uuid4().hex
        role1 = unit.new_role_ref(id=role_id1, name='fake1name')
        role2 = unit.new_role_ref(id=role_id2, name='fake2name')
        PROVIDERS.role_api.create_role(role_id1, role1)
        PROVIDERS.role_api.create_role(role_id2, role2)
        role1['name'] = 'fake2name'
        self.assertRaises(
            exception.Conflict, PROVIDERS.role_api.update_role, role_id1, role1
        )

    def test_role_crud(self):
        role = unit.new_role_ref()
        role_name = role['name']
        PROVIDERS.role_api.create_role(role['id'], role)
        role_ref = PROVIDERS.role_api.get_role(role['id'])
        role_ref_dict = {x: role_ref[x] for x in role_ref}
        self.assertDictEqual(role, role_ref_dict)

        role_ref = PROVIDERS.role_api.get_unique_role_by_name(role_name)
        self.assertEqual(role['id'], role_ref['id'])

        role['name'] = uuid.uuid4().hex
        updated_role_ref = PROVIDERS.role_api.update_role(role['id'], role)
        role_ref = PROVIDERS.role_api.get_role(role['id'])
        role_ref_dict = {x: role_ref[x] for x in role_ref}
        self.assertDictEqual(role, role_ref_dict)
        self.assertDictEqual(role_ref_dict, updated_role_ref)

        PROVIDERS.role_api.delete_role(role['id'])
        self.assertRaises(
            exception.RoleNotFound, PROVIDERS.role_api.get_role, role['id']
        )

    def test_role_crud_without_description(self):
        role = {
            'id': uuid.uuid4().hex,
            'name': uuid.uuid4().hex,
            'domain_id': None,
            'options': {},
        }
        self.role_api.create_role(role['id'], role)
        role_ref = self.role_api.get_role(role['id'])
        role_ref_dict = {x: role_ref[x] for x in role_ref}
        self.assertIsNone(role_ref_dict['description'])
        role_ref_dict.pop('description')
        self.assertDictEqual(role, role_ref_dict)

        role['name'] = uuid.uuid4().hex
        updated_role_ref = self.role_api.update_role(role['id'], role)
        role_ref = self.role_api.get_role(role['id'])
        role_ref_dict = {x: role_ref[x] for x in role_ref}
        self.assertIsNone(updated_role_ref['description'])
        self.assertDictEqual(role_ref_dict, updated_role_ref)

        self.role_api.delete_role(role['id'])
        self.assertRaises(
            exception.RoleNotFound, self.role_api.get_role, role['id']
        )

    def test_update_role_returns_not_found(self):
        role = unit.new_role_ref()
        self.assertRaises(
            exception.RoleNotFound,
            PROVIDERS.role_api.update_role,
            role['id'],
            role,
        )

    def test_list_roles(self):
        roles = PROVIDERS.role_api.list_roles()
        self.assertEqual(len(default_fixtures.ROLES), len(roles))
        role_ids = {role['id'] for role in roles}
        expected_role_ids = {role['id'] for role in default_fixtures.ROLES}
        self.assertEqual(expected_role_ids, role_ids)

    @unit.skip_if_cache_disabled('role')
    def test_cache_layer_role_crud(self):
        role = unit.new_role_ref()
        role_id = role['id']
        # Create role
        PROVIDERS.role_api.create_role(role_id, role)
        role_ref = PROVIDERS.role_api.get_role(role_id)
        updated_role_ref = copy.deepcopy(role_ref)
        updated_role_ref['name'] = uuid.uuid4().hex
        # Update role, bypassing the role api manager
        PROVIDERS.role_api.driver.update_role(role_id, updated_role_ref)
        # Verify get_role still returns old ref
        self.assertDictEqual(role_ref, PROVIDERS.role_api.get_role(role_id))
        # Invalidate Cache
        PROVIDERS.role_api.get_role.invalidate(PROVIDERS.role_api, role_id)
        # Verify get_role returns the new role_ref
        self.assertDictEqual(
            updated_role_ref, PROVIDERS.role_api.get_role(role_id)
        )
        # Update role back to original via the assignment api manager
        PROVIDERS.role_api.update_role(role_id, role_ref)
        # Verify get_role returns the original role ref
        self.assertDictEqual(role_ref, PROVIDERS.role_api.get_role(role_id))
        # Delete role bypassing the role api manager
        PROVIDERS.role_api.driver.delete_role(role_id)
        # Verify get_role still returns the role_ref
        self.assertDictEqual(role_ref, PROVIDERS.role_api.get_role(role_id))
        # Invalidate cache
        PROVIDERS.role_api.get_role.invalidate(PROVIDERS.role_api, role_id)
        # Verify RoleNotFound is now raised
        self.assertRaises(
            exception.RoleNotFound, PROVIDERS.role_api.get_role, role_id
        )
        # recreate role
        PROVIDERS.role_api.create_role(role_id, role)
        PROVIDERS.role_api.get_role(role_id)
        # delete role via the assignment api manager
        PROVIDERS.role_api.delete_role(role_id)
        # verity RoleNotFound is now raised
        self.assertRaises(
            exception.RoleNotFound, PROVIDERS.role_api.get_role, role_id
        )

    def test_create_role_immutable(self):
        role = unit.new_role_ref()
        role_id = role['id']
        role['options'][ro_opt.IMMUTABLE_OPT.option_name] = True
        role_created = PROVIDERS.role_api.create_role(role_id, role)
        role_via_manager = PROVIDERS.role_api.get_role(role_id)
        self.assertTrue('options' in role_created)
        self.assertTrue('options' in role_via_manager)
        self.assertTrue(
            role_via_manager['options'][ro_opt.IMMUTABLE_OPT.option_name]
        )
        self.assertTrue(
            role_created['options'][ro_opt.IMMUTABLE_OPT.option_name]
        )

    def test_cannot_update_immutable_role(self):
        role = unit.new_role_ref()
        role_id = role['id']
        role['options'][ro_opt.IMMUTABLE_OPT.option_name] = True
        PROVIDERS.role_api.create_role(role_id, role)
        update_role = {'name': uuid.uuid4().hex}
        self.assertRaises(
            exception.ResourceUpdateForbidden,
            PROVIDERS.role_api.update_role,
            role_id,
            update_role,
        )

    def test_cannot_update_immutable_role_while_unsetting_immutable(self):
        role = unit.new_role_ref()
        role_id = role['id']
        role['options'][ro_opt.IMMUTABLE_OPT.option_name] = True
        PROVIDERS.role_api.create_role(role_id, role)
        update_role = {
            'name': uuid.uuid4().hex,
            'options': {ro_opt.IMMUTABLE_OPT.option_name: True},
        }
        self.assertRaises(
            exception.ResourceUpdateForbidden,
            PROVIDERS.role_api.update_role,
            role_id,
            update_role,
        )

    def test_cannot_delete_immutable_role(self):
        role = unit.new_role_ref()
        role_id = role['id']
        role['options'][ro_opt.IMMUTABLE_OPT.option_name] = True
        PROVIDERS.role_api.create_role(role_id, role)
        self.assertRaises(
            exception.ResourceDeleteForbidden,
            PROVIDERS.role_api.delete_role,
            role_id,
        )

    def test_update_role_set_immutable(self):
        role = unit.new_role_ref()
        role_id = role['id']
        PROVIDERS.role_api.create_role(role_id, role)
        update_role = {'options': {ro_opt.IMMUTABLE_OPT.option_name: True}}
        role_via_manager = PROVIDERS.role_api.get_role(role_id)
        self.assertTrue('options' in role_via_manager)
        self.assertFalse(
            ro_opt.IMMUTABLE_OPT.option_name in role_via_manager['options']
        )
        role_update = PROVIDERS.role_api.update_role(role_id, update_role)
        role_via_manager = PROVIDERS.role_api.get_role(role_id)
        self.assertTrue(
            ro_opt.IMMUTABLE_OPT.option_name in role_update['options']
        )
        self.assertTrue(
            role_update['options'][ro_opt.IMMUTABLE_OPT.option_name]
        )
        self.assertTrue(
            ro_opt.IMMUTABLE_OPT.option_name in role_via_manager['options']
        )
        self.assertTrue(
            role_via_manager['options'][ro_opt.IMMUTABLE_OPT.option_name]
        )

    def test_update_role_set_immutable_with_additional_updates(self):
        role = unit.new_role_ref()
        role_id = role['id']
        PROVIDERS.role_api.create_role(role_id, role)
        update_role = {
            'name': uuid.uuid4().hex,
            'options': {ro_opt.IMMUTABLE_OPT.option_name: True},
        }
        role_via_manager = PROVIDERS.role_api.get_role(role_id)
        self.assertTrue('options' in role_via_manager)
        self.assertFalse(
            ro_opt.IMMUTABLE_OPT.option_name in role_via_manager['options']
        )
        role_update = PROVIDERS.role_api.update_role(role_id, update_role)
        role_via_manager = PROVIDERS.role_api.get_role(role_id)
        self.assertEqual(role_update['name'], update_role['name'])
        self.assertEqual(role_via_manager['name'], update_role['name'])
        self.assertTrue(
            ro_opt.IMMUTABLE_OPT.option_name in role_update['options']
        )
        self.assertTrue(
            role_update['options'][ro_opt.IMMUTABLE_OPT.option_name]
        )
        self.assertTrue(
            ro_opt.IMMUTABLE_OPT.option_name in role_via_manager['options']
        )
        self.assertTrue(
            role_via_manager['options'][ro_opt.IMMUTABLE_OPT.option_name]
        )

    def test_update_role_unset_immutable(self):
        role = unit.new_role_ref()
        role_id = role['id']
        role['options'][ro_opt.IMMUTABLE_OPT.option_name] = True
        PROVIDERS.role_api.create_role(role_id, role)
        role_via_manager = PROVIDERS.role_api.get_role(role_id)
        self.assertTrue('options' in role_via_manager)
        self.assertTrue(
            role_via_manager['options'][ro_opt.IMMUTABLE_OPT.option_name]
        )
        update_role = {'options': {ro_opt.IMMUTABLE_OPT.option_name: False}}
        PROVIDERS.role_api.update_role(role_id, update_role)
        role_via_manager = PROVIDERS.role_api.get_role(role_id)
        self.assertTrue('options' in role_via_manager)
        self.assertTrue(
            ro_opt.IMMUTABLE_OPT.option_name in role_via_manager['options']
        )
        self.assertFalse(
            role_via_manager['options'][ro_opt.IMMUTABLE_OPT.option_name]
        )
        update_role = {'options': {ro_opt.IMMUTABLE_OPT.option_name: None}}
        role_updated = PROVIDERS.role_api.update_role(role_id, update_role)
        role_via_manager = PROVIDERS.role_api.get_role(role_id)
        self.assertTrue('options' in role_updated)
        self.assertTrue('options' in role_via_manager)
        self.assertFalse(
            ro_opt.IMMUTABLE_OPT.option_name in role_updated['options']
        )
        self.assertFalse(
            ro_opt.IMMUTABLE_OPT.option_name in role_via_manager['options']
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Single point of entry to generate the sample configuration file.

This module collects all the necessary info from the other modules in this
package. It is assumed that:

* Every other module in this package has a 'list_opts' function which
  returns a dict where:

  * The keys are strings which are the group names.

  * The value of each key is a list of config options for that group.

* The conf package doesn't have further packages with config options.

* This module is only used in the context of sample file generation.

"""

import collections
import importlib
import os
import pkgutil

LIST_OPTS_FUNC_NAME = 'list_opts'
IGNORED_MODULES = ('opts', 'constants', 'utils')


def list_opts():
    opts = collections.defaultdict(list)
    module_names = _list_module_names()
    imported_modules = _import_modules(module_names)
    _append_config_options(imported_modules, opts)
    return _tupleize(opts)


def _tupleize(d):
    """Convert a dict of options to the 2-tuple format."""
    return list(d.items())


def _list_module_names():
    module_names = []
    package_path = os.path.dirname(os.path.abspath(__file__))
    for _, module_name, ispkg in pkgutil.iter_modules(path=[package_path]):
        if module_name in IGNORED_MODULES or ispkg:
            # Skip this module.
            continue
        else:
            module_names.append(module_name)
    return module_names


def _import_modules(module_names):
    imported_modules = []
    for module_name in module_names:
        full_module_path = '.'.join(__name__.split('.')[:-1] + [module_name])
        module = importlib.import_module(full_module_path)
        if not hasattr(module, LIST_OPTS_FUNC_NAME):
            raise Exception(
                f"The module '{full_module_path}' should have a '{LIST_OPTS_FUNC_NAME}' function which "
                "returns the config options."
            )
        else:
            imported_modules.append(module)
    return imported_modules


def _append_config_options(imported_modules, config_options):
    for module in imported_modules:
        configs = module.list_opts()
        for key, val in configs.items():
            config_options[key].extend(val)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# This file handles all flask-restful resources for /v3/OS-FEDERATION

import http.client

import flask
import flask_restful
from oslo_log import log
from oslo_serialization import jsonutils

from keystone.api._shared import authentication
from keystone.api._shared import json_home_relations
from keystone.api import validation
from keystone.common import provider_api
from keystone.common import rbac_enforcer
from keystone.common import render_token
from keystone.common import validation as ks_validation
import keystone.conf
from keystone import exception
from keystone.federation import schema
from keystone.federation import utils
from keystone.server import flask as ks_flask

LOG = log.getLogger(__name__)
CONF = keystone.conf.CONF
ENFORCER = rbac_enforcer.RBACEnforcer
PROVIDERS = provider_api.ProviderAPIs


_build_param_relation = json_home_relations.os_federation_parameter_rel_func
_build_resource_relation = json_home_relations.os_federation_resource_rel_func

IDP_ID_PARAMETER_RELATION = _build_param_relation(parameter_name='idp_id')
PROTOCOL_ID_PARAMETER_RELATION = _build_param_relation(
    parameter_name='protocol_id'
)
SERVICE_PROVIDER_ID_PARAMETER_RELATION = _build_param_relation(
    parameter_name='service_provider_id'
)


def _combine_lists_uniquely(a, b):
    # it's most likely that only one of these will be filled so avoid
    # the combination if possible.
    if a and b:
        return {x['id']: x for x in a + b}.values()
    else:
        return a or b


class _ResourceBase(ks_flask.ResourceBase):
    json_home_resource_rel_func = _build_resource_relation
    json_home_parameter_rel_func = _build_param_relation

    @classmethod
    def wrap_member(cls, ref, collection_name=None, member_name=None):
        cls._add_self_referential_link(ref, collection_name)
        cls._add_related_links(ref)
        return {member_name or cls.member_key: cls.filter_params(ref)}

    @staticmethod
    def _add_related_links(ref):
        # Do Nothing, This is in support of child class mechanisms.
        pass


class IdentityProvidersResource(_ResourceBase):
    collection_key = 'identity_providers'
    member_key = 'identity_provider'
    api_prefix = '/OS-FEDERATION'
    _public_parameters = frozenset(
        [
            'id',
            'enabled',
            'description',
            'remote_ids',
            'links',
            'domain_id',
            'authorization_ttl',
        ]
    )
    _id_path_param_name_override = 'idp_id'

    @staticmethod
    def _add_related_links(ref):
        """Add URLs for entities related with Identity Provider.

        Add URLs pointing to:
        - protocols tied to the Identity Provider

        """
        base_path = ref['links'].get('self')
        if base_path is None:
            base_path = '/'.join(
                ks_flask.base_url(path='/{}'.format(ref['id']))
            )

        for name in ['protocols']:
            ref['links'][name] = '/'.join([base_path, name])

    @validation.request_query_schema(
        schema.identity_provider_index_request_query
    )
    @validation.response_body_schema(
        schema.identity_provider_index_response_body
    )
    def get(self):
        """List all identity providers.

        GET/HEAD /OS-FEDERATION/identity_providers
        """
        filters = ['id', 'enabled']
        ENFORCER.enforce_call(
            action='identity:list_identity_providers', filters=filters
        )
        hints = self.build_driver_hints(filters)
        refs = PROVIDERS.federation_api.list_idps(hints=hints)
        refs = [self.filter_params(r) for r in refs]
        collection = self.wrap_collection(refs, hints=hints)
        for r in collection[self.collection_key]:
            # Add the related links explicitly
            self._add_related_links(r)
        return collection


class IdentityProviderResource(_ResourceBase):
    collection_key = 'identity_providers'
    member_key = 'identity_provider'
    api_prefix = '/OS-FEDERATION'
    _public_parameters = frozenset(
        [
            'id',
            'enabled',
            'description',
            'remote_ids',
            'links',
            'domain_id',
            'authorization_ttl',
        ]
    )
    _id_path_param_name_override = 'idp_id'

    @validation.request_query_schema(schema.identity_provider_request_query)
    @validation.response_body_schema(
        schema.identity_provider_index_response_body
    )
    def get(self, idp_id=None):
        """Get an IDP resource.

        GET/HEAD /OS-FEDERATION/identity_providers/{idp_id}
        """
        ENFORCER.enforce_call(action='identity:get_identity_provider')
        ref = PROVIDERS.federation_api.get_idp(idp_id)
        return self.wrap_member(ref)

    @validation.request_body_schema(
        schema.identity_provider_create_request_body
    )
    @validation.response_body_schema(schema.identity_provider_response_body)
    def put(self, idp_id):
        """Create an idp resource for federated authentication.

        PUT /OS-FEDERATION/identity_providers/{idp_id}
        """
        ENFORCER.enforce_call(action='identity:create_identity_provider')
        idp = self.request_body_json.get('identity_provider', {})
        idp = self._normalize_dict(idp)
        idp.setdefault('enabled', False)
        idp_ref = PROVIDERS.federation_api.create_idp(idp_id, idp)
        return self.wrap_member(idp_ref), http.client.CREATED

    @validation.request_body_schema(
        schema.identity_provider_update_request_body
    )
    @validation.response_body_schema(schema.identity_provider_response_body)
    def patch(self, idp_id):
        ENFORCER.enforce_call(action='identity:update_identity_provider')
        idp = self.request_body_json.get('identity_provider', {})
        idp = self._normalize_dict(idp)
        idp_ref = PROVIDERS.federation_api.update_idp(idp_id, idp)
        return self.wrap_member(idp_ref)

    def delete(self, idp_id):
        ENFORCER.enforce_call(action='identity:delete_identity_provider')
        PROVIDERS.federation_api.delete_idp(idp_id)
        return None, http.client.NO_CONTENT


class _IdentityProvidersProtocolsResourceBase(_ResourceBase):
    collection_key = 'protocols'
    member_key = 'protocol'
    _public_parameters = frozenset(['id', 'mapping_id', 'links'])
    json_home_additional_parameters = {'idp_id': IDP_ID_PARAMETER_RELATION}
    json_home_collection_resource_name_override = 'identity_provider_protocols'
    json_home_member_resource_name_override = 'identity_provider_protocol'

    @staticmethod
    def _add_related_links(ref):
        """Add new entries to the 'links' subdictionary in the response.

        Adds 'identity_provider' key with URL pointing to related identity
        provider as a value.

        :param ref: response dictionary

        """
        ref.setdefault('links', {})
        ref['links']['identity_provider'] = ks_flask.base_url(
            path=ref['idp_id']
        )


class IDPProtocolsListResource(_IdentityProvidersProtocolsResourceBase):
    def get(self, idp_id):
        """List protocols for an IDP.

        HEAD/GET /OS-FEDERATION/identity_providers/{idp_id}/protocols
        """
        ENFORCER.enforce_call(action='identity:list_protocols')
        protocol_refs = PROVIDERS.federation_api.list_protocols(idp_id)
        protocols = list(protocol_refs)
        collection = self.wrap_collection(protocols)
        for r in collection[self.collection_key]:
            # explicitly add related links
            self._add_related_links(r)
        return collection


class IDPProtocolsCRUDResource(_IdentityProvidersProtocolsResourceBase):
    def get(self, idp_id, protocol_id):
        """Get protocols for an IDP.

        HEAD/GET /OS-FEDERATION/identity_providers/
                 {idp_id}/protocols/{protocol_id}
        """
        ENFORCER.enforce_call(action='identity:get_protocol')
        ref = PROVIDERS.federation_api.get_protocol(idp_id, protocol_id)
        return self.wrap_member(ref)

    def put(self, idp_id, protocol_id):
        """Create protocol for an IDP.

        PUT /OS-Federation/identity_providers/{idp_id}/protocols/{protocol_id}
        """
        ENFORCER.enforce_call(action='identity:create_protocol')
        protocol = self.request_body_json.get('protocol', {})
        ks_validation.lazy_validate(schema.protocol_create, protocol)
        protocol = self._normalize_dict(protocol)
        ref = PROVIDERS.federation_api.create_protocol(
            idp_id, protocol_id, protocol
        )
        return self.wrap_member(ref), http.client.CREATED

    def patch(self, idp_id, protocol_id):
        """Update protocol for an IDP.

        PATCH /OS-FEDERATION/identity_providers/
              {idp_id}/protocols/{protocol_id}
        """
        ENFORCER.enforce_call(action='identity:update_protocol')
        protocol = self.request_body_json.get('protocol', {})
        ks_validation.lazy_validate(schema.protocol_update, protocol)
        ref = PROVIDERS.federation_api.update_protocol(
            idp_id, protocol_id, protocol
        )
        return self.wrap_member(ref)

    def delete(self, idp_id, protocol_id):
        """Delete a protocol from an IDP.

        DELETE /OS-FEDERATION/identity_providers/
               {idp_id}/protocols/{protocol_id}
        """
        ENFORCER.enforce_call(action='identity:delete_protocol')
        PROVIDERS.federation_api.delete_protocol(idp_id, protocol_id)
        return None, http.client.NO_CONTENT


class MappingResource(_ResourceBase):
    collection_key = 'mappings'
    member_key = 'mapping'
    api_prefix = '/OS-FEDERATION'

    def get(self, mapping_id=None):
        if mapping_id is not None:
            return self._get_mapping(mapping_id)
        return self._list_mappings()

    def _get_mapping(self, mapping_id):
        """Get a mapping.

        HEAD/GET /OS-FEDERATION/mappings/{mapping_id}
        """
        ENFORCER.enforce_call(action='identity:get_mapping')
        return self.wrap_member(
            PROVIDERS.federation_api.get_mapping(mapping_id)
        )

    def _list_mappings(self):
        """List mappings.

        HEAD/GET /OS-FEDERATION/mappings
        """
        ENFORCER.enforce_call(action='identity:list_mappings')
        return self.wrap_collection(PROVIDERS.federation_api.list_mappings())

    def _internal_normalize_and_validate_attribute_mapping(
        self, action_executed_message="created"
    ):
        mapping = self.request_body_json.get('mapping', {})
        mapping = self._normalize_dict(mapping)

        if not mapping.get('schema_version'):
            default_schema_version = (
                utils.get_default_attribute_mapping_schema_version()
            )
            LOG.debug(
                "A mapping [%s] was %s without providing a "
                "'schema_version'; therefore, we need to set one. The "
                "current default is [%s]. We will use this value for "
                "the attribute mapping being registered. It is "
                "recommended that one does not rely on this default "
                "value, as it can change, and the already persisted "
                "attribute mappings will remain with the previous "
                "default values.",
                mapping,
                action_executed_message,
                default_schema_version,
            )
            mapping['schema_version'] = default_schema_version
        utils.validate_mapping_structure(mapping)
        return mapping

    def put(self, mapping_id):
        """Create a mapping.

        PUT /OS-FEDERATION/mappings/{mapping_id}
        """
        ENFORCER.enforce_call(action='identity:create_mapping')

        am = self._internal_normalize_and_validate_attribute_mapping(
            "registered"
        )
        mapping_ref = PROVIDERS.federation_api.create_mapping(mapping_id, am)

        return self.wrap_member(mapping_ref), http.client.CREATED

    def patch(self, mapping_id):
        """Update an attribute mapping for identity federation.

        PATCH /OS-FEDERATION/mappings/{mapping_id}
        """
        ENFORCER.enforce_call(action='identity:update_mapping')

        am = self._internal_normalize_and_validate_attribute_mapping("updated")
        mapping_ref = PROVIDERS.federation_api.update_mapping(mapping_id, am)

        return self.wrap_member(mapping_ref)

    def delete(self, mapping_id):
        """Delete a mapping.

        DELETE /OS-FEDERATION/mappings/{mapping_id}
        """
        ENFORCER.enforce_call(action='identity:delete_mapping')
        PROVIDERS.federation_api.delete_mapping(mapping_id)
        return None, http.client.NO_CONTENT


class ServiceProvidersResource(_ResourceBase):
    collection_key = 'service_providers'
    member_key = 'service_provider'
    _public_parameters = frozenset(
        [
            'auth_url',
            'id',
            'enabled',
            'description',
            'links',
            'relay_state_prefix',
            'sp_url',
        ]
    )
    api_prefix = '/OS-FEDERATION'

    @validation.request_query_schema(
        schema.service_provider_index_request_query
    )
    @validation.response_body_schema(
        schema.service_provider_index_response_body
    )
    def get(self):
        """List service providers.

        GET/HEAD /OS-FEDERATION/service_providers
        """
        filters = ['id', 'enabled']
        ENFORCER.enforce_call(
            action='identity:list_service_providers', filters=filters
        )
        hints = self.build_driver_hints(filters)
        refs = [
            self.filter_params(r)
            for r in PROVIDERS.federation_api.list_sps(hints=hints)
        ]
        return self.wrap_collection(refs, hints=hints)


class ServiceProviderResource(_ResourceBase):
    collection_key = 'service_providers'
    member_key = 'service_provider'
    _public_parameters = frozenset(
        [
            'auth_url',
            'id',
            'enabled',
            'description',
            'links',
            'relay_state_prefix',
            'sp_url',
        ]
    )
    api_prefix = '/OS-FEDERATION'

    @validation.request_query_schema(None)
    @validation.response_body_schema(schema.service_provider_response_body)
    def get(self, service_provider_id):
        """Get a service provider.

        GET/HEAD /OS-FEDERATION/service_providers/{service_provider_id}
        """
        ENFORCER.enforce_call(action='identity:get_service_provider')
        return self.wrap_member(
            PROVIDERS.federation_api.get_sp(service_provider_id)
        )

    @validation.request_body_schema(
        schema.service_provider_create_request_body
    )
    @validation.response_body_schema(schema.service_provider_response_body)
    def put(self, service_provider_id):
        """Create a service provider.

        PUT /OS-FEDERATION/service_providers/{service_provider_id}
        """
        ENFORCER.enforce_call(action='identity:create_service_provider')
        sp = self.request_body_json.get('service_provider', {})
        sp = self._normalize_dict(sp)
        sp.setdefault('enabled', False)
        sp.setdefault('relay_state_prefix', CONF.saml.relay_state_prefix)
        sp_ref = PROVIDERS.federation_api.create_sp(service_provider_id, sp)
        return self.wrap_member(sp_ref), http.client.CREATED

    @validation.request_body_schema(
        schema.service_provider_update_request_body
    )
    @validation.response_body_schema(schema.service_provider_response_body)
    def patch(self, service_provider_id):
        """Update a service provider.

        PATCH /OS-FEDERATION/service_providers/{service_provider_id}
        """
        ENFORCER.enforce_call(action='identity:update_service_provider')
        sp = self.request_body_json.get('service_provider', {})
        sp = self._normalize_dict(sp)
        sp_ref = PROVIDERS.federation_api.update_sp(service_provider_id, sp)
        return self.wrap_member(sp_ref)

    def delete(self, service_provider_id):
        """Delete a service provider.

        DELETE /OS-FEDERATION/service_providers/{service_provider_id}
        """
        ENFORCER.enforce_call(action='identity:delete_service_provider')
        PROVIDERS.federation_api.delete_sp(service_provider_id)
        return None, http.client.NO_CONTENT


class SAML2MetadataResource(flask_restful.Resource):
    @ks_flask.unenforced_api
    def get(self):
        """Get SAML2 metadata.

        GET/HEAD /OS-FEDERATION/saml2/metadata
        """
        metadata_path = CONF.saml.idp_metadata_path
        try:
            with open(metadata_path) as metadata_handler:
                metadata = metadata_handler.read()
        except OSError as e:
            # Raise HTTP 500 in case Metadata file cannot be read.
            raise exception.MetadataFileError(reason=e)
        resp = flask.make_response(metadata, http.client.OK)
        resp.headers['Content-Type'] = 'text/xml'
        return resp


class OSFederationAuthResource(flask_restful.Resource):
    @ks_flask.unenforced_api
    def get(self, idp_id, protocol_id):
        """Authenticate from dedicated uri endpoint.

        GET/HEAD /OS-FEDERATION/identity_providers/
                 {idp_id}/protocols/{protocol_id}/auth
        """
        return self._auth(idp_id, protocol_id)

    @ks_flask.unenforced_api
    def post(self, idp_id, protocol_id):
        """Authenticate from dedicated uri endpoint.

        POST /OS-FEDERATION/identity_providers/
             {idp_id}/protocols/{protocol_id}/auth
        """
        return self._auth(idp_id, protocol_id)

    def _auth(self, idp_id, protocol_id):
        """Build and pass auth data to authentication code.

        Build HTTP request body for federated authentication and inject
        it into the ``authenticate_for_token`` function.
        """
        auth = {
            'identity': {
                'methods': [protocol_id],
                protocol_id: {
                    'identity_provider': idp_id,
                    'protocol': protocol_id,
                },
            }
        }
        token = authentication.authenticate_for_token(auth)
        token_data = render_token.render_token_response_from_model(token)
        resp_data = jsonutils.dumps(token_data)
        flask_resp = flask.make_response(resp_data, http.client.CREATED)
        flask_resp.headers['X-Subject-Token'] = token.id
        flask_resp.headers['Content-Type'] = 'application/json'
        return flask_resp


class OSFederationAPI(ks_flask.APIBase):
    _name = 'OS-FEDERATION'
    _import_name = __name__
    _api_url_prefix = '/OS-FEDERATION'
    resources = []
    resource_mapping = [
        ks_flask.construct_resource_map(
            resource=SAML2MetadataResource,
            url='/saml2/metadata',
            resource_kwargs={},
            rel='metadata',
            resource_relation_func=_build_resource_relation,
        ),
        ks_flask.construct_resource_map(
            resource=OSFederationAuthResource,
            url=(
                '/identity_providers/<string:idp_id>/protocols/'
                '<string:protocol_id>/auth'
            ),
            resource_kwargs={},
            rel='identity_provider_protocol_auth',
            resource_relation_func=_build_resource_relation,
            path_vars={
                'idp_id': IDP_ID_PARAMETER_RELATION,
                'protocol_id': PROTOCOL_ID_PARAMETER_RELATION,
            },
        ),
    ]


class OSFederationIdentityProvidersAPI(ks_flask.APIBase):
    _name = 'identity_providers'
    _import_name = __name__
    _api_url_prefix = '/OS-FEDERATION'
    resources = []
    resource_mapping = [
        ks_flask.construct_resource_map(
            resource=IdentityProvidersResource,
            url='/identity_providers',
            resource_kwargs={},
            rel="identity_providers",
            resource_relation_func=_build_resource_relation,
            path_vars=None,
        ),
        ks_flask.construct_resource_map(
            resource=IdentityProviderResource,
            url='/identity_providers/<string:idp_id>',
            resource_kwargs={},
            rel="identity_provider",
            resource_relation_func=_build_resource_relation,
            path_vars={'idp_id': IDP_ID_PARAMETER_RELATION},
        ),
    ]


class OSFederationIdentityProvidersProtocolsAPI(ks_flask.APIBase):
    _name = 'protocols'
    _import_name = __name__
    resources = []
    resource_mapping = [
        ks_flask.construct_resource_map(
            resource=IDPProtocolsCRUDResource,
            url=(
                '/OS-FEDERATION/identity_providers/<string:idp_id>/protocols/'
                '<string:protocol_id>'
            ),
            resource_kwargs={},
            rel='identity_provider_protocol',
            resource_relation_func=_build_resource_relation,
            path_vars={
                'idp_id': IDP_ID_PARAMETER_RELATION,
                'protocol_id': PROTOCOL_ID_PARAMETER_RELATION,
            },
        ),
        ks_flask.construct_resource_map(
            resource=IDPProtocolsListResource,
            url='/OS-FEDERATION/identity_providers/<string:idp_id>/protocols',
            resource_kwargs={},
            rel='identity_provider_protocols',
            resource_relation_func=_build_resource_relation,
            path_vars={'idp_id': IDP_ID_PARAMETER_RELATION},
        ),
    ]


class OSFederationMappingsAPI(ks_flask.APIBase):
    _name = 'mappings'
    _import_name = __name__
    _api_url_prefix = '/OS-FEDERATION'
    resources = [MappingResource]
    resource_mapping = []


class OSFederationServiceProvidersAPI(ks_flask.APIBase):
    _name = 'service_providers'
    _import_name = __name__
    _api_url_prefix = '/OS-FEDERATION'
    resource_mapping = [
        ks_flask.construct_resource_map(
            resource=ServiceProvidersResource,
            url='/service_providers',
            resource_kwargs={},
            rel="service_providers",
            resource_relation_func=_build_resource_relation,
            path_vars=None,
        ),
        ks_flask.construct_resource_map(
            resource=ServiceProviderResource,
            url='/service_providers/<string:service_provider_id>',
            resource_kwargs={},
            rel="service_provider",
            resource_relation_func=_build_resource_relation,
            path_vars={
                'service_provider_id': SERVICE_PROVIDER_ID_PARAMETER_RELATION
            },
        ),
    ]


APIs = (
    OSFederationAPI,
    OSFederationIdentityProvidersAPI,
    OSFederationIdentityProvidersProtocolsAPI,
    OSFederationMappingsAPI,
    OSFederationServiceProvidersAPI,
)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Unified in-memory receipt model."""

from oslo_log import log
from oslo_serialization import msgpackutils
from oslo_utils import reflection

from keystone.auth import core
from keystone.common import cache
from keystone.common import provider_api
from keystone import exception
from keystone.identity.backends import resource_options as ro

LOG = log.getLogger(__name__)
PROVIDERS = provider_api.ProviderAPIs


class ReceiptModel:
    """An object that represents a receipt emitted by keystone.

    This is a queryable object that other parts of keystone can use to reason
    about a user's receipt.
    """

    def __init__(self):
        self.user_id = None
        self.__user = None
        self.__user_domain = None

        self.methods = None
        self.__required_methods = None

        self.__expires_at = None
        self.__issued_at = None

    def __repr__(self):
        """Return string representation of KeystoneReceipt."""
        desc = '<%(type)s at %(loc)s>'
        self_cls_name = reflection.get_class_name(self, fully_qualified=False)
        return desc % {'type': self_cls_name, 'loc': hex(id(self))}

    @property
    def expires_at(self):
        return self.__expires_at

    @expires_at.setter
    def expires_at(self, value):
        if not isinstance(value, str):
            raise ValueError('expires_at must be a string.')
        self.__expires_at = value

    @property
    def issued_at(self):
        return self.__issued_at

    @issued_at.setter
    def issued_at(self, value):
        if not isinstance(value, str):
            raise ValueError('issued_at must be a string.')
        self.__issued_at = value

    @property
    def user(self):
        if not self.__user:
            if self.user_id:
                self.__user = PROVIDERS.identity_api.get_user(self.user_id)
        return self.__user

    @property
    def user_domain(self):
        if not self.__user_domain:
            if self.user:
                self.__user_domain = PROVIDERS.resource_api.get_domain(
                    self.user['domain_id']
                )
        return self.__user_domain

    @property
    def required_methods(self):
        if not self.__required_methods:
            mfa_rules = self.user['options'].get(
                ro.MFA_RULES_OPT.option_name, []
            )
            rules = core.UserMFARulesValidator._parse_rule_structure(
                mfa_rules, self.user_id
            )
            methods = set(self.methods)
            active_methods = set(core.AUTH_METHODS.keys())

            required_auth_methods = []
            for r in rules:
                r_set = set(r).intersection(active_methods)
                if r_set.intersection(methods):
                    required_auth_methods.append(list(r_set))

            self.__required_methods = required_auth_methods

        return self.__required_methods

    def mint(self, receipt_id, issued_at):
        """Set the ``id`` and ``issued_at`` attributes of a receipt.

        The process of building a Receipt requires setting attributes about the
        partial authentication context, like ``user_id`` and ``methods`` for
        example. Once a Receipt object accurately represents this information
        it should be "minted". Receipt are minted when they get an ``id``
        attribute and their creation time is recorded.
        """
        self.id = receipt_id
        self.issued_at = issued_at


class _ReceiptModelHandler:
    identity = 125
    handles = (ReceiptModel,)

    def __init__(self, registry):
        self._registry = registry

    def serialize(self, obj):
        serialized = msgpackutils.dumps(obj.__dict__, registry=self._registry)
        return serialized

    def deserialize(self, data):
        receipt_data = msgpackutils.loads(data, registry=self._registry)
        try:
            receipt_model = ReceiptModel()
            for k, v in iter(receipt_data.items()):
                setattr(receipt_model, k, v)
        except Exception:
            LOG.debug(
                "Failed to deserialize ReceiptModel. Data is %s", receipt_data
            )
            raise exception.CacheDeserializationError(
                ReceiptModel.__name__, receipt_data
            )
        return receipt_model


cache.register_model_handler(_ReceiptModelHandler)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc


class Provider(metaclass=abc.ABCMeta):
    """Interface for credential providers that support encryption."""

    @abc.abstractmethod
    def encrypt(self, credential):
        """Encrypt a credential.

        :param str credential: credential to encrypt
        :returns: encrypted credential str
        :raises: keystone.exception.CredentialEncryptionError
        """

    @abc.abstractmethod
    def decrypt(self, credential):
        """Decrypt a credential.

        :param str credential: credential to decrypt
        :returns: credential str as plaintext
        :raises: keystone.exception.CredentialEncryptionError
        """



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# This file handles all flask-restful resources for /v3/ec2tokens

import http.client
import urllib.parse

import flask
from keystoneclient.contrib.ec2 import utils as ec2_utils
from oslo_serialization import jsonutils

from keystone.api._shared import EC2_S3_Resource
from keystone.api._shared import json_home_relations
from keystone.common import render_token
from keystone.common import utils
from keystone import exception
from keystone.i18n import _
from keystone.server import flask as ks_flask

CRED_TYPE_EC2 = 'ec2'


class EC2TokensResource(EC2_S3_Resource.ResourceBase):
    @staticmethod
    def _check_signature(creds_ref, credentials):
        signer = ec2_utils.Ec2Signer(creds_ref['secret'])
        signature = signer.generate(credentials)
        # NOTE(davechecn): credentials.get('signature') is not guaranteed to
        # exist, we need to check it explicitly.
        if credentials.get('signature'):
            if utils.auth_str_equal(credentials['signature'], signature):
                return True
            # NOTE(vish): Some client libraries don't use the port when
            # signing requests, so try again without the port.
            elif ':' in credentials['host']:
                parsed = urllib.parse.urlsplit('//' + credentials['host'])
                credentials['host'] = parsed.hostname
                # NOTE(davechen): we need to reinitialize 'signer' to avoid
                # contaminated status of signature, this is similar with
                # other programming language libraries, JAVA for example.
                signer = ec2_utils.Ec2Signer(creds_ref['secret'])
                signature = signer.generate(credentials)
                if utils.auth_str_equal(credentials['signature'], signature):
                    return True
            raise exception.Unauthorized(_('Invalid EC2 signature.'))
        # Raise the exception when credentials.get('signature') is None
        else:
            raise exception.Unauthorized(_('EC2 signature not supplied.'))

    @ks_flask.unenforced_api
    def post(self):
        """Authenticate ec2 token.

        POST /v3/ec2tokens
        """
        token = self.handle_authenticate()
        token_reference = render_token.render_token_response_from_model(token)
        resp_body = jsonutils.dumps(token_reference)
        response = flask.make_response(resp_body, http.client.OK)
        response.headers['X-Subject-Token'] = token.id
        response.headers['Content-Type'] = 'application/json'
        return response


class EC2TokensAPI(ks_flask.APIBase):
    _name = 'ec2tokens'
    _import_name = __name__
    resources = []
    resource_mapping = [
        ks_flask.construct_resource_map(
            resource=EC2TokensResource,
            url='/ec2tokens',
            resource_kwargs={},
            rel='ec2tokens',
            resource_relation_func=(
                json_home_relations.os_ec2_resource_rel_func
            ),
        )
    ]


APIs = (EC2TokensAPI,)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


class ProviderAPIRegistry:
    __shared_object_state: dict = {}
    __registry: dict = {}
    __iter__ = __registry.__iter__
    __getitem__ = __registry.__getitem__
    locked = False

    def __init__(self):
        # NOTE(morgan): This rebinds __dict__ and allows all instances of
        # the provider API to share a common state. Any changes except
        # rebinding __dict__ will maintain the same state stored on the class
        # not the instance. This design pattern is preferable to
        # full singletons where state sharing is the important "feature"
        # derived from the "singleton"
        #
        # Use "super" to bypass the __setattr__ preventing changes to the
        # object itself.
        super().__setattr__('__dict__', self.__shared_object_state)

    def __getattr__(self, item):
        """Do attr lookup."""
        try:
            return self.__registry[item]
        except KeyError:
            raise AttributeError(f"'ProviderAPIs' has no attribute {item}")

    def __setattr__(self, key, value):
        """Do not allow setting values on the registry object."""
        raise RuntimeError(
            'Programming Error: You may not set values on the '
            'ProviderAPIRegistry objects.'
        )

    def _register_provider_api(self, name, obj):
        """Register an instance of a class as a provider api."""
        if name == 'driver':
            raise ValueError('A provider may not be named "driver".')

        if self.locked:
            raise RuntimeError(
                'Programming Error: The provider api registry has been '
                'locked (post configuration). Ensure all provider api '
                'managers are instantiated before locking.'
            )

        if name in self.__registry:
            raise DuplicateProviderError(
                f'`{name}` has already been registered as an api '
                f'provider by `{self.__registry[name]!r}`'
            )
        self.__registry[name] = obj

    def _clear_registry_instances(self):
        """ONLY USED FOR TESTING."""
        self.__registry.clear()
        # Use super to allow setting around class implementation of __setattr__
        super().__setattr__('locked', False)

    def lock_provider_registry(self):
        # Use super to allow setting around class implementation of __setattr__
        super().__setattr__('locked', True)

    def deferred_provider_lookup(self, api, method):
        """Create descriptor that performs lookup of api and method on demand.

        For specialized cases, such as the enforcer "get_member_from_driver"
        which needs to be effectively a "classmethod", this method returns
        a smart descriptor object that does the lookup at runtime instead of
        at import time.

        :param api: The api to use, e.g. "identity_api"
        :type api: str
        :param method: the method on the api to return
        :type method: str
        """

        class DeferredProviderLookup:
            def __init__(self, api, method):
                self.__api = api
                self.__method = method

            def __get__(self, instance, owner):
                api = getattr(ProviderAPIs, self.__api)
                return getattr(api, self.__method)

        return DeferredProviderLookup(api, method)


class DuplicateProviderError(Exception):
    """Attempting to register a duplicate API provider."""


class ProviderAPIMixin:
    """Allow referencing provider apis on self via __getattr__.

    Be sure this class is first in the class definition for inheritance.
    """

    def __getattr__(self, item):
        """Magic getattr method."""
        try:
            return getattr(ProviderAPIs, item)
        except AttributeError:
            return self.__getattribute__(item)


ProviderAPIs = ProviderAPIRegistry()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import http.client
import uuid

from keystone.common import provider_api
import keystone.conf
from keystone.tests.common import auth as common_auth
from keystone.tests import unit
from keystone.tests.unit import base_classes
from keystone.tests.unit import ksfixtures

CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs


class _SystemUserOauth1ConsumerTests:
    """Common default functionality for all system users."""

    def test_user_can_get_consumer(self):
        ref = PROVIDERS.oauth_api.create_consumer({'id': uuid.uuid4().hex})
        with self.test_client() as c:
            c.get(
                '/v3/OS-OAUTH1/consumers/{}'.format(ref['id']),
                headers=self.headers,
            )

    def test_user_can_list_consumers(self):
        PROVIDERS.oauth_api.create_consumer({'id': uuid.uuid4().hex})
        with self.test_client() as c:
            c.get('/v3/OS-OAUTH1/consumers', headers=self.headers)


class _SystemReaderAndMemberOauth1ConsumerTests:
    def test_user_cannot_create_consumer(self):
        with self.test_client() as c:
            c.post(
                '/v3/OS-OAUTH1/consumers',
                json={'consumer': {}},
                expected_status_code=http.client.FORBIDDEN,
                headers=self.headers,
            )

    def test_user_cannot_update_consumer(self):
        ref = PROVIDERS.oauth_api.create_consumer({'id': uuid.uuid4().hex})
        with self.test_client() as c:
            c.patch(
                '/v3/OS-OAUTH1/consumers/{}'.format(ref['id']),
                json={'consumer': {'description': uuid.uuid4().hex}},
                expected_status_code=http.client.FORBIDDEN,
                headers=self.headers,
            )

    def test_user_cannot_delete_consumer(self):
        ref = PROVIDERS.oauth_api.create_consumer({'id': uuid.uuid4().hex})
        with self.test_client() as c:
            c.delete(
                '/v3/OS-OAUTH1/consumers/{}'.format(ref['id']),
                expected_status_code=http.client.FORBIDDEN,
                headers=self.headers,
            )


class SystemReaderTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _SystemUserOauth1ConsumerTests,
    _SystemReaderAndMemberOauth1ConsumerTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        system_reader = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(system_reader)['id']
        PROVIDERS.assignment_api.create_system_grant_for_user(
            self.user_id, self.bootstrapper.reader_role_id
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=system_reader['password'],
            system=True,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class SystemMemberTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _SystemUserOauth1ConsumerTests,
    _SystemReaderAndMemberOauth1ConsumerTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        system_member = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(system_member)['id']
        PROVIDERS.assignment_api.create_system_grant_for_user(
            self.user_id, self.bootstrapper.member_role_id
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=system_member['password'],
            system=True,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class SystemAdminTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _SystemUserOauth1ConsumerTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        # Reuse the system administrator account created during
        # ``keystone-manage bootstrap``
        self.user_id = self.bootstrapper.admin_user_id
        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=self.bootstrapper.admin_password,
            system=True,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}

    def test_user_can_create_consumer(self):
        with self.test_client() as c:
            c.post(
                '/v3/OS-OAUTH1/consumers',
                json={'consumer': {}},
                headers=self.headers,
            )

    def test_user_can_update_consumer(self):
        ref = PROVIDERS.oauth_api.create_consumer({'id': uuid.uuid4().hex})
        with self.test_client() as c:
            c.patch(
                '/v3/OS-OAUTH1/consumers/{}'.format(ref['id']),
                json={'consumer': {'description': uuid.uuid4().hex}},
                headers=self.headers,
            )

    def test_user_can_delete_consumer(self):
        ref = PROVIDERS.oauth_api.create_consumer({'id': uuid.uuid4().hex})
        with self.test_client() as c:
            c.delete(
                '/v3/OS-OAUTH1/consumers/{}'.format(ref['id']),
                headers=self.headers,
            )
